/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___StatusUpdateEvent___
#define ___StatusUpdateEvent___

#include <wx/event.h>
#include <wx/stopwatch.h>

#if wxVERSION_NUMBER < 2900
#	define wxEVENT_HANDLER_CAST( functype, func)	(wxObjectEventFunction)(wxEventFunction)wxStaticCastEvent( functype, & func)
#endif

namespace ProceduralTextures
{
	extern const wxEventType wxEVT_STATUS_UPDATE;
	//! Event to manage the update of FPS count
	/*!
	This event is lauched after each draw to tell the main window that a frame was drawn, so it can update the FPS count
	\author Sylvain DOREMUS
	\version 0.6.1.0
	\date 01/04/2011
	*/
	class wxStatusUpdateEvent : public wxNotifyEvent
	{
	private:
		wxMilliClock_t m_savedTime;
		wxMilliClock_t m_time;
		int m_iIndex;

	public:
		wxStatusUpdateEvent( wxEventType p_eventType, int p_iIndex=0, int p_winid=wxID_ANY);
		wxStatusUpdateEvent( const wxStatusUpdateEvent & p_event);

		virtual wxEvent * Clone() const;
		void ResetTime();

		inline wxMilliClock_t	GetTime		()const	{ return m_time; }
		inline int				GetIndex	()const	{ return m_iIndex; }

//		DECLARE_DYNAMIC_CLASS( wxStatusUpdateEvent)
	};


	typedef void (wxEvtHandler:: * wxStatusUpdateEventFunction)( wxStatusUpdateEvent &);
#	define wxStatusUpdateEventHandler( func)	wxEVENT_HANDLER_CAST( wxStatusUpdateEventFunction, func)
#	define EVT_STATUS_UPDATE( id, fn)			DECLARE_EVENT_TABLE_ENTRY( wxEVT_STATUS_UPDATE, id, wxID_ANY, wxStatusUpdateEventHandler( fn), (wxObject *) NULL),

#if defined( __WXMSW__ ) && wxVERSION_NUMBER < 2900
	extern const wxEventType wxEVT_THREAD;

	class wxThreadEvent : public wxEvent
	{
	private:
		wxString m_strString;

	public:
		wxThreadEvent( wxEventType p_eventType = wxEVT_THREAD, int p_id = wxID_ANY);
		wxThreadEvent(const wxThreadEvent & p_event);

		virtual wxEvent * Clone()const;

		inline const wxString	&	GetString	()const { return m_strString; }
		inline void SetString	( const wxString & val)	{ m_strString = val; }

		DECLARE_DYNAMIC_CLASS( wxThreadEvent)
	};

	typedef void (wxEvtHandler:: * wxThreadEventFunction)( wxThreadEvent &);
#	define wxThreadEventHandler( func)	wxEVENT_HANDLER_CAST( wxThreadEventFunction, func)
#	define EVT_THREAD( id, fn)			DECLARE_EVENT_TABLE_ENTRY( wxEVT_THREAD, id, wxID_ANY, wxThreadEventHandler( & fn), (wxObject *) NULL),
#endif
}
#endif
