/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___PerlinNoise___
#define ___PerlinNoise___

#include <LibGenerator/ProceduralGenerator.h>

namespace ProceduralTextures
{
	class PerlinNoise : public ProceduralGenerator
	{
	public:
		typedef enum
		{	eNOISE_TYPE_TURBULENCE
		,	eNOISE_TYPE_CLOUD
		,	eNOISE_TYPE_MARBLE
		,	eNOISE_TYPE_PARAM
		,	eNOISE_TYPE_COUNT
		}	eNOISE_TYPE;

		typedef enum
		{	eNOISE_PARAM_FUNCTION_COS
		,	eNOISE_PARAM_FUNCTION_SIN
		,	eNOISE_PARAM_FUNCTION_TAN
		,	eNOISE_PARAM_FUNCTION_COSH
		,	eNOISE_PARAM_FUNCTION_SINH
		,	eNOISE_PARAM_FUNCTION_TANH
		,	eNOISE_PARAM_FUNCTION_ACOS
		,	eNOISE_PARAM_FUNCTION_ASIN
		,	eNOISE_PARAM_FUNCTION_ATAN
		,	eNOISE_PARAM_FUNCTION_SQRT
		,	eNOISE_PARAM_FUNCTION_LN
		,	eNOISE_PARAM_FUNCTION_EXP
		,	eNOISE_PARAM_FUNCTION_LOG10
		,	eNOISE_PARAM_FUNCTION_COUNT
		}	eNOISE_PARAM_FUNCTION;

		static wxString FuncNames[eNOISE_PARAM_FUNCTION_COUNT];
		static wxString TypeNames[eNOISE_TYPE_COUNT];

	private:
		typedef float (* NoiseFunction)( float);

		class Thread : public ProceduralGenerator::Thread
		{
		private:
			typedef void (Thread:: * TypeFunction)( int x, int y, int z, UbPixel & p_rPixel);

		public:
			int m_iPermutations[512];
			eNOISE_TYPE m_eType;
			int m_iStep;
			int m_iNumberOfOctaves;
			float m_fPersistency;
			PixelBuffer * m_pixels;
			NoiseFunction m_pfnNoise;
			TypeFunction m_pfnType;
			TypeFunction m_pfnArrayTypes[eNOISE_TYPE_COUNT];

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour);
			virtual ~Thread();

			virtual void Step();

			void SetType( eNOISE_TYPE p_eType);
			void SetFunction( NoiseFunction p_pfnFunction);

			inline void SetBuffer( PixelBuffer * p_pixels) { m_pixels = p_pixels; }

		private:
			void _turbulence	( int x, int y, int z, UbPixel & p_rPixel);
			void _clouds		( int x, int y, int z, UbPixel & p_rPixel);
			void _marble		( int x, int y, int z, UbPixel & p_rPixel);
			void _param			( int x, int y, int z, UbPixel & p_rPixel);
			double _fade( double t);
			double _lerp( double t, double a, double b);
			double _grad(int hash, double x, double y, double z);
			double _noise(double x, double y, double z);
			void _perlin();
		};

	private:
		int m_iNumberOfOctaves;
		float m_fPersistency;
		int m_iPermutations[512];
		int m_iStep;
		eNOISE_TYPE m_eType;
		NoiseFunction m_functions[eNOISE_PARAM_FUNCTION_COUNT];
		bool m_bModifying;

	public:
		PerlinNoise( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame, int p_iNumberOfOctaves=2, float p_fPersistency=0.05f, eNOISE_TYPE p_eType=eNOISE_TYPE_TURBULENCE);
		virtual ~PerlinNoise();

		void SetNumberOfOctaves( int val);
		void SetPersistency( float val);
		void SetType( eNOISE_TYPE p_eType);
		void SetFunction( eNOISE_PARAM_FUNCTION p_eFunction);

		int		GetNumberOfOctaves	()const { return m_iNumberOfOctaves; }
		float	GetPersistency		()const { return m_fPersistency; }

	private:
		virtual void _glInit		(){}
		virtual void _glPreRender	(){}
		virtual void _glRender		( bool & p_bChanged){}
		virtual void _glPostRender	(){}
		virtual void _glCleanup		(){}
	};

	class PnPanel : public ConfigPanel
	{
	private:
		typedef enum
		{	eRedValue	= 50
		,	eGreenValue	= 51
		,	eBlueValue	= 52
		,	eOctaves	= 53
		,	eFrequency	= 54
		,	eType		= 55
		,	eFunction	= 56
		}	eIDs;

		PerlinNoise		*	m_pGenerator;
		wxSlider		*	m_pRedValue;
		wxStaticText	*	m_pRedText;
		wxSlider		*	m_pGreenValue;
		wxStaticText	*	m_pGreenText;
		wxSlider		*	m_pBlueValue;
		wxStaticText	*	m_pBlueText;
		wxStaticText	*	m_pStaticOctaves;
		wxStaticText	*	m_pStaticFrequency;
		wxSlider		*	m_pSliderOctaves;
		wxSlider		*	m_pSliderFrequency;
		wxComboBox		*	m_pComboType;
		wxComboBox		*	m_pComboFunction;

	public:
		PnPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
		virtual ~PnPanel();

		void Create( PerlinNoise * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size);

	private:
		DECLARE_EVENT_TABLE()
		void _onSliderRed		( wxCommandEvent	&	p_event);
		void _onSliderGreen		( wxCommandEvent	&	p_event);
		void _onSliderBlue		( wxCommandEvent	&	p_event);
		void _onOctaves			( wxCommandEvent	&	p_event);
		void _onFrequency		( wxCommandEvent	&	p_event);
		void _onType			( wxCommandEvent	&	p_event);
		void _onFunction		( wxCommandEvent	&	p_event);
	};

	class PnPlugin : public Plugin<PnPanel, PerlinNoise, PnPlugin>
	{
	public:
#ifdef __WXMSW__
		PnPlugin( wxEventType p_eventType=wxEVT_STATUS_UPDATE );
#else
		PnPlugin( wxEventType p_eventType );
#endif
		virtual ~PnPlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
