/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___MainFrame___
#define ___MainFrame___

namespace ProceduralTextures
{
    class RenderPanel;
	class StatusBar;

	class MainFrame : public wxFrame
	{
	private:
		RenderPanel						*	m_pRenderPanel;
		ConfigPanel						*	m_pConfigPanel;
		std::map <wxString, PluginBase *>	m_mapPluginsByName;
		PluginBase						*	m_pPlugin;
		StatusBar						*	m_pStatusBar;
		static wxEventType 					m_eventType;

	public:
		enum
		{	eMyID				= 1
		,	eSelectGenerator	= 2
		,	ePrintScreen		= 3
		,	eExit				= 4
		};

		MainFrame();
		~MainFrame();

		void Initialise();

	private:
		DECLARE_CLASS( MainFrame);
		void _buildMenuBar();
		ProceduralGenerator * _selectGenerator();
		ProceduralGenerator * _loadGenerator( const wxString & p_strName, const wxSize & p_size);
		void _destroyGenerator();
		void _loadPlugins();
		void _close();
		void DoLoadPlugins( wxString const & p_strPath );

		DECLARE_EVENT_TABLE();
		void _onPaint			( wxPaintEvent			& 	p_event );
		void _onSize			( wxSizeEvent			&	p_event );
		void _onClose			( wxCloseEvent			&	p_event );
		void _onStatusUpdate	( wxNotifyEvent			&	p_event );
		void _onUpdateFps		( wxCommandEvent		&	p_event );
		void _onKeyUp			( wxKeyEvent			&	p_event );
		void _onExit			( wxCommandEvent		&	p_event );
		void _onSelectGenerator	( wxCommandEvent		&	p_event );
		void _onPrintScreen		( wxCommandEvent		&	p_event );
	};
}

#endif
