/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___Pg_RenderPanel___
#define ___Pg_RenderPanel___

#include <wx/timer.h>
#include <wx/glcanvas.h>

namespace ProceduralTextures
{
	class ProceduralGenerator;
	class wxStatusUpdateEvent;

	class RenderPanel : public wxGLCanvas
	{
	private:
		static const int		m_iTimerId = 27;
		static wxString			m_strGlErrors[6];
		ProceduralGenerator	*	m_pGenerator;
		wxTimer				*	m_pTimer;
		wxGLContext			*	m_pContext;
		bool					m_bInitialised;
		wxStatusUpdateEvent	*	m_pEvent;

	public:
		RenderPanel( wxEventType p_eventType, wxWindow * p_parent, int * p_attribList, const wxPoint & p_position, const wxSize & p_size);
		~RenderPanel();

		void Resize( const wxSize & p_size);
		void SetGenerator( ProceduralGenerator * p_pGenerator);
		void SaveFrame();

	private:
		void _render( bool p_bChanged);
		void _initGL();

		static void _drawElements( void * p_pThis);

		DECLARE_EVENT_TABLE();
		void _onPaint( wxPaintEvent & p_event);
		void _onSize( wxSizeEvent & p_event);
		void _onTimer( wxTimerEvent & p_event);
		void _onClose( wxCloseEvent & p_event);
		void _onKeyUp( wxKeyEvent & p_event);
	};
}

#endif
