#include "BuggedLifeGame/GeneratorBuggedLifeGame.h"
#include <LibGenerator/Random.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

BlgCell :: BlgCell()
	:	m_alivePx( NULL)
	,	m_deadPx( NULL)
	,	m_nextPixel( NULL)
{
}

BlgCell :: BlgCell( const BlgCell & p_cell)
	:	m_alivePx( p_cell.m_alivePx)
	,	m_deadPx( p_cell.m_deadPx)
	,	m_nextPixel( p_cell.m_nextPixel)
	,	m_pixel( p_cell.m_pixel)
	,	m_step( p_cell.m_step)
	,	m_age( p_cell.m_age)
	,	m_med( p_cell.m_med)
{
	for (size_t i = 0 ; i < 8 ; i++)
	{
		m_neighbours[i] = p_cell.m_neighbours[i];
	}
}

BlgCell :: ~BlgCell()
{
}

BlgCell & BlgCell :: operator =( const BlgCell & p_cell)
{
	m_alivePx = p_cell.m_alivePx;
	m_deadPx = p_cell.m_deadPx;
	for (size_t i = 0 ; i < 8 ; i++)
	{
		m_neighbours[i] = p_cell.m_neighbours[i];
	}
	m_alive = p_cell.m_alive;
	m_nextPixel = p_cell.m_nextPixel;
	m_pixel = p_cell.m_pixel;
	m_step = p_cell.m_step;
	m_age = p_cell.m_age;
	m_med = p_cell.m_med;
	return * this;
}

void BlgCell :: Set( UbPixel * p_pixel, UbPixel * p_alivePx, UbPixel * p_deadPx, bool p_alive)
{
	m_alivePx = p_alivePx;
	m_deadPx = p_deadPx;
	m_nextPixel = p_pixel;

	if( m_alivePx && m_deadPx && m_nextPixel )
	{
		m_step.r = ((unsigned int)(m_alivePx->r) + 1 - m_deadPx->r) >> 5;
		m_step.g = ((unsigned int)(m_alivePx->g) + 1 - m_deadPx->g) >> 5;
		m_step.b = ((unsigned int)(m_alivePx->b) + 1 - m_deadPx->b) >> 5;
		m_step.a = ((unsigned int)(m_alivePx->a) + 1 - m_deadPx->a) >> 5;

		SetAlive( p_alive);
	}
}

void BlgCell :: Die()
{
	m_alive = false;

	*m_nextPixel = Substract( m_pixel, m_step, *m_deadPx);
	m_med = *m_nextPixel / m_step;
}

void BlgCell :: Live()
{
	if ( ! m_alive)
	{
		m_age.r = 0;
		m_age.g = 0;
		m_age.b = 0;
		m_age.a = 0;
	}

	m_alive = true;

	m_age += m_step;
	*m_nextPixel = Add( m_pixel, m_age, *m_alivePx);
	m_med = *m_nextPixel / m_step;
}

void BlgCell :: Stay()
{
	if (m_alive)
	{
		Live();
	}
	else
	{
		Die();
	}
}

void BlgCell :: SetAlive( bool p_alive)
{
	m_alive = p_alive;

	if (m_alive)
	{
		*m_nextPixel = *m_alivePx;
	}
	else
	{
		*m_nextPixel = *m_deadPx;
	}
}

void BlgCell :: Update()
{
	m_pixel.r = 0;
	m_pixel.g = 0;
	m_pixel.b = 0;
	m_pixel.a = 0;

	for (int i = 0 ; i < 8 ; i++)
	{
		m_pixel += m_neighbours[i]->m_med;
	}
}

//*************************************************************************************************

BuggedLifeGame::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	ProceduralGenerator::Thread( p_pParent, p_uiIndex, iWidth, iTop, iBottom, iTotalHeight, p_pxColour),
		m_cells( NULL)
{
}

BuggedLifeGame::Thread :: ~Thread()
{
}

void BuggedLifeGame::Thread :: Step()
{
	int l_nbAlive = 0;

	for( int i = 0 ; i < m_iWidth ; i++ )
	{
		for( int j = m_iTop ; j < m_iBottom ; j++ )
		{
			l_nbAlive = _getAliveNeightbours( i, j );

			if( l_nbAlive == 3 )
			{
				(*m_cells)[j][i].Live();
			}
			else if( l_nbAlive != 2 )
			{
				(*m_cells)[j][i].Die();
			}
			else
			{
				(*m_cells)[j][i].Stay();
			}
		}
	}
}

int BuggedLifeGame::Thread :: _getAliveNeightbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_iWidth || p_y < 0 || p_y >= m_iHeight)
	{
		return 0;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_iWidth - 1;
	}
	else if (p_x == m_iWidth - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_iHeight - 1;
	}
	else if (p_y == m_iHeight - 1)
	{
		l_down = 0;
	}

	return (	( *m_cells)[l_up	][l_left	].m_alive ? 1 : 0) + (	( *m_cells)[p_y		][l_left	].m_alive ? 1 : 0) + (	( *m_cells)[l_down	][l_left	].m_alive ? 1 : 0)
		   + (	( *m_cells)[l_up	][p_x		].m_alive ? 1 : 0) + (	( *m_cells)[l_down	][p_x		].m_alive ? 1 : 0)
		   + (	( *m_cells)[l_up	][l_right	].m_alive ? 1 : 0) + (	( *m_cells)[p_y		][l_right	].m_alive ? 1 : 0) + (	( *m_cells)[l_down	][l_right	].m_alive ? 1 : 0);
}

//*************************************************************************************************

BuggedLifeGame :: BuggedLifeGame( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_eventType, p_width, p_height, p_iWndId, p_pFrame)
	,	m_cells( Point<size_t, 2>( p_width, p_height))
	,	m_oldCells( Point<size_t, 2>( p_width, p_height))
{
	m_bNeedsConfigPanel = true;
	Random::srand ( time( NULL));

	m_alivePx = m_pxColour;

	m_deadPx.r = 0;
	m_deadPx.g = 0;
	m_deadPx.b = 0;
	m_deadPx.a = 0;

	bool l_alive;

	for (int i = 0 ; i < p_height ; i++)
	{
		for (int j = 0 ; j < p_width ; j++)
		{
			l_alive =  Random::rand() % 2 == 0;
			m_cells[i][j].Set( 		&m_frontBuffer[i][j], 	&m_alivePx, &m_deadPx, l_alive );
			m_oldCells[i][j].Set( 	&m_backBuffer[i][j], 	&m_alivePx, &m_deadPx, l_alive );
		}
	}

	for (int i = 0 ; i < p_width ; i++)
	{
		for (int j = 0 ; j < p_height ; j++)
		{
			DoInitialiseNeighbours( i, j);
		}
	}

	BuggedLifeGame::Thread * l_pThread;
	int l_iOffset = p_height / m_uiThreadCount;

	for (int i = 0 ; i < p_height ; i += l_iOffset)
	{
		l_pThread = CreateThread<BuggedLifeGame::Thread>( p_width, i, i + l_iOffset, p_height, m_pxColour);
		l_pThread->Run();
	}

	m_initialised = true;
}

BuggedLifeGame :: ~BuggedLifeGame()
{
	m_initialised = false;

	_cleanup();

	m_cells.clear();
	m_oldCells.clear();
}

void BuggedLifeGame :: InitialiseStep()
{
	for (int i = 0 ; i < m_sizeImage.x ; i++)
	{
		for (int j = 0 ; j < m_sizeImage.y ; j++)
		{
			m_oldCells[j][i].Update();
		}
	}

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			static_cast<BuggedLifeGame::Thread *>( m_arraySlaveThreads[i])->SetBuffer( & m_oldCells);
			m_arraySlaveThreads[i]->Relaunch();
		}
	}
}

void BuggedLifeGame :: DoInitialiseNeighbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_sizeImage.x || p_y < 0 || p_y >= m_sizeImage.y)
	{
		return;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_sizeImage.x-1;
	}
	else if (p_x == m_sizeImage.x - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_sizeImage.y-1;
	}
	else if (p_y == m_sizeImage.y - 1)
	{
		l_down = 0;
	}

	m_cells[p_y][p_x].m_neighbours[0] = & m_cells[l_up		][l_left	];
	m_cells[p_y][p_x].m_neighbours[1] = & m_cells[p_y		][l_left	];
	m_cells[p_y][p_x].m_neighbours[2] = & m_cells[l_down	][l_left	];
	m_cells[p_y][p_x].m_neighbours[3] = & m_cells[l_up		][p_x		];
	m_cells[p_y][p_x].m_neighbours[4] = & m_cells[l_down	][p_x		];
	m_cells[p_y][p_x].m_neighbours[5] = & m_cells[l_up		][l_right	];
	m_cells[p_y][p_x].m_neighbours[6] = & m_cells[p_y		][l_right	];
	m_cells[p_y][p_x].m_neighbours[7] = & m_cells[l_down	][l_right	];

	m_oldCells[p_y][p_x].m_neighbours[0] = & m_oldCells[l_up	][l_left	];
	m_oldCells[p_y][p_x].m_neighbours[1] = & m_oldCells[p_y		][l_left	];
	m_oldCells[p_y][p_x].m_neighbours[2] = & m_oldCells[l_down	][l_left	];
	m_oldCells[p_y][p_x].m_neighbours[3] = & m_oldCells[l_up	][p_x		];
	m_oldCells[p_y][p_x].m_neighbours[4] = & m_oldCells[l_down	][p_x		];
	m_oldCells[p_y][p_x].m_neighbours[5] = & m_oldCells[l_up	][l_right	];
	m_oldCells[p_y][p_x].m_neighbours[6] = & m_oldCells[p_y		][l_right	];
	m_oldCells[p_y][p_x].m_neighbours[7] = & m_oldCells[l_down	][l_right	];
}

int BuggedLifeGame :: _getAliveNeightbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_sizeImage.x || p_y < 0 || p_y >= m_sizeImage.y)
	{
		return 0;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_sizeImage.x - 1;
	}
	else if (p_x == m_sizeImage.x - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_sizeImage.y - 1;
	}
	else if (p_y == m_sizeImage.y - 1)
	{
		l_down = 0;
	}

	return (	m_cells[l_up	][l_left	].m_alive ? 1 : 0) + (	m_cells[p_y		][l_left	].m_alive ? 1 : 0) + (	m_cells[l_down	][l_left	].m_alive ? 1 : 0)
		   + (	m_cells[l_up	][p_x		].m_alive ? 1 : 0) + (	m_cells[l_down	][p_x		].m_alive ? 1 : 0)
		   + (	m_cells[l_up	][l_right	].m_alive ? 1 : 0) + (	m_cells[p_y		][l_right	].m_alive ? 1 : 0) + (	m_cells[l_down	][l_right	].m_alive ? 1 : 0);
}

//******************************************************************************

IMPLEMENT_PLUGIN( BlgPlugin, BlgPanel, BuggedLifeGame)

BlgPlugin :: BlgPlugin( wxEventType p_eventType )
	:	Plugin< BlgPanel, BuggedLifeGame, BlgPlugin >( p_eventType )
{
}

BlgPlugin :: ~BlgPlugin()
{
}

wxString BlgPlugin :: GetName()
{
#if defined( NDEBUG )
	return wxT( "Bugged Life Game");
#else
	return wxT( "Bugged Life Game (Debug)");
#endif
}

//*************************************************************************************************

BlgPanel :: BlgPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
{
}

BlgPanel :: ~BlgPanel()
{
}

void BlgPanel :: Create( BuggedLifeGame * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	Reparent( p_pWindow);
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	m_pGenerator = p_pGenerator;

	wxString l_strValue;
	l_strValue << int( m_pGenerator->GetRed());
	new wxStaticText( this, wxID_ANY, wxT( "R"), wxPoint( 20, 0), wxSize( 20, 16));
	m_pRedText = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 20, 16), wxSize( 20, 16));
	m_pRedValue = new wxSlider( this, eRedValue, m_pGenerator->GetRed(), 0, 255, wxPoint( 20, 32), wxSize( 20, 60), wxSL_VERTICAL);

	l_strValue.clear();
	l_strValue << int( m_pGenerator->GetGreen());
	new wxStaticText( this, wxID_ANY, wxT( "G"), wxPoint( 60, 0), wxSize( 20, 16));
	m_pGreenText = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 60, 16), wxSize( 20, 16));
	m_pGreenValue = new wxSlider( this, eGreenValue, m_pGenerator->GetGreen(), 0, 255, wxPoint( 60, 32), wxSize( 20, 60), wxSL_VERTICAL);

	l_strValue.clear();
	l_strValue << int( m_pGenerator->GetBlue());
	new wxStaticText( this, wxID_ANY, wxT( "B"), wxPoint( 100, 0), wxSize( 20, 16));
	m_pBlueText = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 100, 16), wxSize( 20, 16));
	m_pBlueValue = new wxSlider( this, eBlueValue, m_pGenerator->GetBlue(), 0, 255, wxPoint( 100, 32), wxSize( 20, 60), wxSL_VERTICAL);
}

BEGIN_EVENT_TABLE( BlgPanel, ConfigPanel)
	EVT_SLIDER( eRedValue,		BlgPanel::_onSliderRed)
	EVT_SLIDER( eGreenValue,	BlgPanel::_onSliderGreen)
	EVT_SLIDER( eBlueValue,		BlgPanel::_onSliderBlue)
END_EVENT_TABLE()

void BlgPanel :: _onSliderRed( wxCommandEvent & p_event)
{
	m_pGenerator->SetRed( p_event.GetInt());
	wxString l_strValue;
	l_strValue << p_event.GetInt();
	m_pRedText->SetLabel( l_strValue);
}

void BlgPanel :: _onSliderGreen( wxCommandEvent & p_event)
{
	m_pGenerator->SetGreen( p_event.GetInt());
	wxString l_strValue;
	l_strValue << p_event.GetInt();
	m_pGreenText->SetLabel( l_strValue);
}

void BlgPanel :: _onSliderBlue( wxCommandEvent & p_event)
{
	m_pGenerator->SetBlue( p_event.GetInt());
	wxString l_strValue;
	l_strValue << p_event.GetInt();
	m_pBlueText->SetLabel( l_strValue);
}

//*************************************************************************************************
