#include "ColoredLifeGame/GeneratorColoredLifeGame.h"
#include <LibGenerator/Random.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

ClgCell :: ClgCell()
	:	m_nextPixel		( NULL)
	,	m_deadPixel		( NULL)
	,	m_alivePixel	( NULL)
	,	m_stepPixel		( NULL)
{
}

ClgCell :: ClgCell( const ClgCell & cell)
	:	m_nextPixel		( cell.m_nextPixel)
	,	m_deadPixel		( cell.m_deadPixel)
	,	m_alivePixel	( cell.m_alivePixel)
	,	m_stepPixel		( cell.m_stepPixel)
	,	m_alive			( cell.m_alive)
{
}

ClgCell :: ~ClgCell()
{
}

ClgCell &  ClgCell :: operator =( const ClgCell & cell)
{
	m_nextPixel		= cell.m_nextPixel;
	m_deadPixel		= cell.m_deadPixel;
	m_alivePixel	= cell.m_alivePixel;
	m_stepPixel		= cell.m_stepPixel;
	m_alive			= cell.m_alive;

	return * this;
}

void ClgCell :: Set( UbPixel * p_pixel, UbPixel * p_deadPixel, UbPixel * p_alivePixel, UiPixel * p_stepPixel, bool p_alive)
{
	m_nextPixel		= p_pixel;
	m_deadPixel		= p_deadPixel;
	m_alivePixel	= p_alivePixel;
	m_stepPixel		= p_stepPixel;
	m_alive			= p_alive;

	if (m_alivePixel != NULL && m_deadPixel != NULL && m_nextPixel != NULL && m_stepPixel != NULL)
	{
		SetAlive( p_alive);

		if (m_alive)
		{
			m_nextPixel->r = m_alivePixel->r;
			m_nextPixel->g = m_alivePixel->g;
			m_nextPixel->b = m_alivePixel->b;
			m_nextPixel->a = m_alivePixel->a;
		}
		else
		{
			m_nextPixel->r = m_deadPixel->r;
			m_nextPixel->g = m_deadPixel->g;
			m_nextPixel->b = m_deadPixel->b;
			m_nextPixel->a = m_deadPixel->a;
		}
	}
}

void ClgCell :: Die()
{
	m_alive = false;
	*m_nextPixel = Add( *m_nextPixel, *m_stepPixel, *m_deadPixel);
}

void ClgCell :: Live()
{
	m_alive = true;

	m_nextPixel->r = m_alivePixel->r;
	m_nextPixel->g = m_alivePixel->g;
	m_nextPixel->b = m_alivePixel->b;
	m_nextPixel->a = m_alivePixel->a;
}

void ClgCell :: SetAlive( bool p_alive)
{
	bool l_alive = m_alive;

	if (p_alive)
	{
		Live();
	}
	else
	{
		Die();
	}
}

//*************************************************************************************************

ColoredLifeGame::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	ProceduralGenerator::Thread( p_pParent, p_uiIndex, iWidth, iTop, iBottom, iTotalHeight, p_pxColour)
	,	m_cells( NULL)
	,	m_oldCells( NULL)
{
}

ColoredLifeGame::Thread :: ~Thread()
{
}

void ColoredLifeGame::Thread :: Step()
{
	int l_iAlive;

	for (int i = 0 ; i < m_iWidth ; i++)
	{
		for (int j = m_iTop ; j < m_iBottom ; j++)
		{
			l_iAlive = _getAliveNeightbours( i, j);

			if (l_iAlive == 3)
			{
				(*m_cells)[j][i].Live();
			}
			else if (l_iAlive != 2)
			{
				(*m_cells)[j][i].Die();
			}
			else
			{
				(*m_cells)[j][i].SetAlive( (*m_oldCells)[j][i].m_alive);
			}
		}
	}
}

int ColoredLifeGame::Thread :: _getAliveNeightbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_iWidth || p_y < 0 || p_y >= m_iHeight)
	{
		return 0;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_iWidth-1;
	}
	else if (p_x == m_iWidth - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_iHeight-1;
	}
	else if (p_y == m_iHeight - 1)
	{
		l_down = 0;
	}

	int l_iResult = 0;

	if ((*m_oldCells)[l_up][l_left].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[p_y][l_left].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_down][l_left].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_up][p_x].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_down][p_x].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_up][l_right].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[p_y][l_right].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_down][l_right].m_alive)
	{
		l_iResult++;
	}

	return l_iResult;
}

//*************************************************************************************************

ColoredLifeGame :: ColoredLifeGame( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_eventType, p_width, p_height, p_iWndId, p_pFrame)
	,	m_oldCells( Point<int, 2>( p_width, p_height))
	,	m_cells( Point<int, 2>( p_width, p_height))
{
	m_bNeedsConfigPanel = true;
	m_pxColour.r = 0;
	m_pxColour.g = 0;
	m_pxColour.b = 0;
	Random::srand ( time( NULL));

	m_alivePixel = m_pxColour;

	m_deadPixel.r = 255;
	m_deadPixel.g = 255;
	m_deadPixel.b = 255;
	m_deadPixel.a = 255;

	m_stepPixel.r = abs( int(static_cast<unsigned int>( m_deadPixel.r) + 1 - static_cast<unsigned int>( m_alivePixel.r)) >> 4);
	m_stepPixel.g = abs( int(static_cast<unsigned int>( m_deadPixel.g) + 1 - static_cast<unsigned int>( m_alivePixel.g)) >> 4);
	m_stepPixel.b = abs( int(static_cast<unsigned int>( m_deadPixel.b) + 1 - static_cast<unsigned int>( m_alivePixel.b)) >> 4);
	m_stepPixel.a = abs( int(static_cast<unsigned int>( m_deadPixel.a) + 1 - static_cast<unsigned int>( m_alivePixel.a)) >> 4);

	bool l_alive;

	for (int i = 0 ; i < p_height ; i++)
	{
		for (int j = 0 ; j < p_width ; j++)
		{
			l_alive =  Random::rand() % 2 == 0;
			m_cells[i][j].Set( & m_frontBuffer[i][j], & m_deadPixel, & m_alivePixel, & m_stepPixel, l_alive);
			m_oldCells[i][j].Set( & m_backBuffer[i][j], & m_deadPixel, & m_alivePixel, & m_stepPixel, l_alive);
		}
	}

	ColoredLifeGame::Thread * l_pThread;
	int l_iOffset = p_height / m_uiThreadCount;

	for (int i = 0 ; i < p_height ; i += l_iOffset)
	{
		l_pThread = CreateThread<ColoredLifeGame::Thread>( p_width, i, i + l_iOffset, p_height, m_pxColour);
		l_pThread->SetCells( & m_cells, & m_oldCells);
		l_pThread->Run();
	}

	m_initialised = true;
}

ColoredLifeGame :: ~ColoredLifeGame()
{
	m_initialised = false;

	_cleanup();

	m_cells.clear();
	m_oldCells.clear();
}

void ColoredLifeGame :: SwapBuffers()
{
	m_oldCells.swap( m_cells);
}

void ColoredLifeGame :: SetRed( unsigned char val)
{
	m_pxColour.r = val;
	m_alivePixel.r = val;

	m_stepPixel.r = abs( int(static_cast<unsigned int>( m_deadPixel.r) + 1 - static_cast<unsigned int>( m_alivePixel.r)) >> 4);
}
void ColoredLifeGame :: SetGreen( unsigned char val)
{
	m_pxColour.g = val;
	m_alivePixel.g = val;

	m_stepPixel.g = abs( int(static_cast<unsigned int>( m_deadPixel.g) + 1 - static_cast<unsigned int>( m_alivePixel.g)) >> 4);
}

void ColoredLifeGame :: SetBlue( unsigned char val)
{
	m_pxColour.b = val;
	m_alivePixel.b = val;

	m_stepPixel.b = abs( int(static_cast<unsigned int>( m_deadPixel.b) + 1 - static_cast<unsigned int>( m_alivePixel.b)) >> 4);
}

//******************************************************************************

IMPLEMENT_PLUGIN( ClgPlugin, ClgPanel, ColoredLifeGame)

ClgPlugin :: ClgPlugin( wxEventType p_eventType )
	:	Plugin< ClgPanel, ColoredLifeGame, ClgPlugin >( p_eventType )
{
}

ClgPlugin :: ~ClgPlugin()
{
}

wxString ClgPlugin :: GetName()
{
#if defined( NDEBUG )
	return wxT( "Coloured Life Game");
#else
	return wxT( "Coloured Life Game (Debug)");
#endif
}

//*************************************************************************************************

ClgPanel :: ClgPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
{
}

ClgPanel :: ~ClgPanel()
{
}

void ClgPanel :: Create( ColoredLifeGame * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	Reparent( p_pWindow);
	m_pGenerator = p_pGenerator;

	wxString l_strValue;
	l_strValue << int( m_pGenerator->GetRed());
	new wxStaticText( this, wxID_ANY, wxT( "R"), wxPoint( 20, 0), wxSize( 20, 16));
	m_pRedText = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 20, 16), wxSize( 20, 16));
	m_pRedValue = new wxSlider( this, eRedValue, m_pGenerator->GetRed(), 0, 255, wxPoint( 20, 32), wxSize( 20, 60), wxSL_VERTICAL);

	l_strValue.clear();
	l_strValue << int( m_pGenerator->GetGreen());
	new wxStaticText( this, wxID_ANY, wxT( "G"), wxPoint( 60, 0), wxSize( 20, 16));
	m_pGreenText = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 60, 16), wxSize( 20, 16));
	m_pGreenValue = new wxSlider( this, eGreenValue, m_pGenerator->GetGreen(), 0, 255, wxPoint( 60, 32), wxSize( 20, 60), wxSL_VERTICAL);

	l_strValue.clear();
	l_strValue << int( m_pGenerator->GetBlue());
	new wxStaticText( this, wxID_ANY, wxT( "B"), wxPoint( 100, 0), wxSize( 20, 16));
	m_pBlueText = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 100, 16), wxSize( 20, 16));
	m_pBlueValue = new wxSlider( this, eBlueValue, m_pGenerator->GetBlue(), 0, 255, wxPoint( 100, 32), wxSize( 20, 60), wxSL_VERTICAL);
}

BEGIN_EVENT_TABLE( ClgPanel, ConfigPanel)
	EVT_SLIDER( eRedValue,		ClgPanel::_onSliderRed)
	EVT_SLIDER( eGreenValue,	ClgPanel::_onSliderGreen)
	EVT_SLIDER( eBlueValue,		ClgPanel::_onSliderBlue)
END_EVENT_TABLE()

void ClgPanel :: _onSliderRed( wxCommandEvent & p_event)
{
	m_pGenerator->SetRed( p_event.GetInt());
	wxString l_strValue;
	l_strValue << p_event.GetInt();
	m_pRedText->SetLabel( l_strValue);
}

void ClgPanel :: _onSliderGreen( wxCommandEvent & p_event)
{
	m_pGenerator->SetGreen( p_event.GetInt());
	wxString l_strValue;
	l_strValue << p_event.GetInt();
	m_pGreenText->SetLabel( l_strValue);
}

void ClgPanel :: _onSliderBlue( wxCommandEvent & p_event)
{
	m_pGenerator->SetBlue( p_event.GetInt());
	wxString l_strValue;
	l_strValue << p_event.GetInt();
	m_pBlueText->SetLabel( l_strValue);
}

//*************************************************************************************************
