#include "LibGenerator/PrecompiledHeader.h"

#include "LibGenerator/BufferElement.h"

#ifdef __WXMSW__
//#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

void ProceduralTextures::Initialise( BufferElementDeclaration & p_declaration, unsigned int p_uiStream, eELEMENT_USAGE p_eUsage, eELEMENT_TYPE p_eDataType, size_t p_uiOffset )
{
	p_declaration.m_uiStream	= p_uiStream;
	p_declaration.m_eUsage		= p_eUsage;
	p_declaration.m_eDataType	= p_eDataType;
	p_declaration.m_uiOffset	= p_uiOffset;
}

//*************************************************************************************************

BufferDeclaration :: BufferDeclaration( const BufferElementDeclaration * p_pElements, size_t p_uiNbElements)
{
	DoInitialise( p_pElements, p_uiNbElements);
}

BufferDeclaration :: ~BufferDeclaration()
{
}

void BufferDeclaration :: DoInitialise( const BufferElementDeclaration * p_pElements, size_t p_uiNbElements)
{
	static const unsigned int s_uiFloat = 4;
	static const unsigned int s_uiColor = 4;
	static const unsigned int s_uiSize[] =
	{	s_uiFloat * 1
	,	s_uiFloat * 2
	,	s_uiFloat * 3
	,	s_uiFloat * 4
	,	s_uiColor * 1
	};
	m_uiStride = 0;

	for (size_t i = 0 ; i < p_uiNbElements ; i++)
	{
		m_arrayElements.push_back( p_pElements[i]);
		m_arrayElements[i].m_uiOffset = m_uiStride;
		m_uiStride += s_uiSize[m_arrayElements[i].m_eDataType];
	}
}

//*************************************************************************************************

BufferElement :: BufferElement( BufferDeclaration * p_pDeclaration )
	:	m_bOwnBuffer	( false )
	,	m_pBuffer		( NULL )
	,	m_uiStride		( 0 )
	,	m_pDeclaration	( p_pDeclaration )
{
	if (m_pDeclaration->Size() > 0)
	{
		size_t i = 0;

		for (BufferDeclaration::const_iterator l_it = m_pDeclaration->Begin() ; l_it != m_pDeclaration->End() ; ++l_it)
		{
			switch (l_it->m_eDataType)
			{
			case eType1Float:
				m_arrayElements.push_back( new Point1f());
				break;

			case eType2Floats:
				m_arrayElements.push_back( new Point2f());
				break;

			case eType3Floats:
				m_arrayElements.push_back( new Point3f());
				break;

			case eType4Floats:
				m_arrayElements.push_back( new Point4f());
				break;

			case eTypeColour:
				m_arrayElements.push_back( new Point4ub());
				break;

			default:
				break;
			}

			m_uiStride += m_arrayElements[i]->GetElementSize() * m_arrayElements[i]->GetElementCount();
			i++;
		}
	}

	UnlinkCoords();
}

BufferElement :: BufferElement( const BufferElement & p_element)
	:	m_bOwnBuffer( false)
	,	m_pBuffer( NULL )
	,	m_uiStride( p_element.m_uiStride)
	,	m_pDeclaration( p_element.m_pDeclaration)
{
	if (m_pDeclaration->Size() > 0)
	{
		size_t i = 0;

		for (BufferDeclaration::const_iterator l_it = m_pDeclaration->Begin() ; l_it != m_pDeclaration->End() ; ++l_it)
		{
			switch (l_it->m_eDataType)
			{
			case eType1Float:
				{
					m_arrayElements.push_back( new Point1f( p_element.GetElement<Point1f>( i)));
				}
				break;

			case eType2Floats:
				{
					m_arrayElements.push_back( new Point2f( p_element.GetElement<Point2f>( i)));
				}
				break;

			case eType3Floats:
				{
					m_arrayElements.push_back( new Point3f( p_element.GetElement<Point3f>( i)));
				}
				break;

			case eType4Floats:
				{
					m_arrayElements.push_back( new Point4f( p_element.GetElement<Point4f>( i)));
				}
				break;

			case eTypeColour:
				{
					m_arrayElements.push_back( new Point4ub( p_element.GetElement<Point4ub>( i)));
				}
				break;

			default:
				break;
			}

			i++;
		}
	}

	UnlinkCoords();
}

BufferElement :: ~BufferElement()
{
	for (size_t i = 0 ; i < m_arrayElements.size() ; i++)
	{
		delete m_arrayElements[i];
	}

	m_arrayElements.clear();

	if (m_bOwnBuffer && m_pBuffer)
	{
		delete [] m_pBuffer;
		m_pBuffer = NULL;
	}
}

BufferElement & BufferElement :: operator =( const BufferElement & p_element)
{
	m_pDeclaration = p_element.m_pDeclaration;

	if (m_pDeclaration->Size() > 0)
	{
		size_t i = 0;

		for (BufferDeclaration::const_iterator l_it = m_pDeclaration->Begin() ; l_it != m_pDeclaration->End() ; ++l_it)
		{
			m_arrayElements[i]->UnlinkCoords();

			switch (l_it->m_eDataType)
			{
			case eType1Float:
				(*m_arrayElements[i]) = p_element.GetElement<Point1f>( i);
				break;

			case eType2Floats:
				(*m_arrayElements[i]) = p_element.GetElement<Point2f>( i);
				break;

			case eType3Floats:
				(*m_arrayElements[i]) = p_element.GetElement<Point3f>( i);
				break;

			case eType4Floats:
				(*m_arrayElements[i]) = p_element.GetElement<Point4f>( i);
				break;

			case eTypeColour:
				(*m_arrayElements[i]) = p_element.GetElement<Point4ub>( i);
				break;

			default:
				break;
			}

			i++;
		}
	}

	m_uiStride = p_element.m_uiStride;

	if (m_bOwnBuffer && m_pBuffer)
	{
		delete [] m_pBuffer;
		m_pBuffer = NULL;
	}

	m_bOwnBuffer = false;
	UnlinkCoords();

	return * this;
}

void BufferElement :: LinkCoords( void * p_pBuffer)
{
	std::vector<unsigned char> l_pBufferSave( m_uiStride);

	if (m_pBuffer)
	{
		memcpy( & l_pBufferSave[0], m_pBuffer, m_uiStride);
	}
	else
	{
		size_t l_uiOffset = 0;

		for (size_t i = 0 ; i < m_arrayElements.size() ; i++)
		{
			memcpy( & l_pBufferSave[l_uiOffset], m_arrayElements[i]->const_ptr<unsigned char>(), m_arrayElements[i]->GetElementSize() * m_arrayElements[i]->GetElementCount());
			l_uiOffset += m_arrayElements[i]->GetElementSize() * m_arrayElements[i]->GetElementCount();
		}
	}

	for (size_t i = 0 ; i < m_arrayElements.size() ; i++)
	{
		m_arrayElements[i]->UnlinkCoords();
	}

	if (m_bOwnBuffer && m_pBuffer)
	{
		delete [] m_pBuffer;
		m_pBuffer = NULL;
	}

	m_pBuffer = (unsigned char *)p_pBuffer;
	m_bOwnBuffer = false;
	DoLink();
	memcpy( m_pBuffer, & l_pBufferSave[0], m_uiStride);
}

void BufferElement :: UnlinkCoords()
{
	std::vector<unsigned char> l_pBufferSave( m_uiStride);

	if (m_pBuffer)
	{
		memcpy( & l_pBufferSave[0], m_pBuffer, m_uiStride);
	}
	else
	{
		size_t l_uiOffset = 0;

		for (size_t i = 0 ; i < m_arrayElements.size() ; i++)
		{
			memcpy( & l_pBufferSave[l_uiOffset], m_arrayElements[i]->const_ptr<unsigned char>(), m_arrayElements[i]->GetElementSize() * m_arrayElements[i]->GetElementCount());
			l_uiOffset += m_arrayElements[i]->GetElementSize() * m_arrayElements[i]->GetElementCount();
		}
	}

	for (size_t i = 0 ; i < m_arrayElements.size() ; i++)
	{
		m_arrayElements[i]->UnlinkCoords();
	}

	if ( ! m_bOwnBuffer)
	{
		m_pBuffer = new unsigned char[m_uiStride];
	}

	m_bOwnBuffer = true;
	DoLink();
	memcpy( m_pBuffer, & l_pBufferSave[0], m_uiStride);
}

void BufferElement :: DoLink()
{
	size_t l_uiOffset = 0;

	for (size_t i = 0 ; i < m_arrayElements.size() ; i++)
	{
		m_arrayElements[i]->LinkCoords( & m_pBuffer[l_uiOffset]);
		l_uiOffset += (m_arrayElements[i]->GetElementSize() * m_arrayElements[i]->GetElementCount());
	}
}

//*************************************************************************************************
