#include "LibGenerator/PrecompiledHeader.h"

#include "LibGenerator/GlObjects.h"
#include "LibGenerator/Vertex.h"
#include "LibGenerator/ProceduralGenerator.h"

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

bool CheckGlError( const wxString & p_strMsg)
{
	static const wxString l_strErrors[] =
	{	wxT( "GL_INVALID_ENUM")
	,	wxT( "GL_INVALID_VALUE")
	,	wxT( "GL_INVALID_OPERATION")
	,	wxT( "GL_STACK_OVERFLOW")
	,	wxT( "GL_STACK_UNDERFLOW")
	,	wxT( "GL_OUT_OF_MEMORY")
	};
	GLenum l_error = OpenGl::GetError();
	bool l_bReturn = true;

	if (l_error != GL_NO_ERROR)
	{
		l_bReturn = false;
	}

	return l_bReturn;
}

//*************************************************************************************************

OpenGl * OpenGl :: m_instance = NULL;

OpenGl :: OpenGl()
	:	m_pfnGetError				( NULL)
	,	m_pfnClearColor				( NULL)
	,	m_pfnShadeModel				( NULL)
	,	m_pfnViewport				( NULL)
	,	m_pfnMatrixMode				( NULL)
	,	m_pfnLoadIdentity			( NULL)
	,	m_pfnOrtho					( NULL)
	,	m_pfnClear					( NULL)
	,	m_pfnDrawElements			( NULL)
	,	m_pfnEnable					( NULL)
	,	m_pfnDisable				( NULL)
	,	m_pfnGenTextures			( NULL)
	,	m_pfnDeleteTextures			( NULL)
	,	m_pfnBindTexture			( NULL)
	,	m_pfnTexSubImage2D			( NULL)
	,	m_pfnTexImage2D				( NULL)
	,	m_pfnTexParameteri			( NULL)
	,	m_pfnGenerateMipmap			( NULL)
	,	m_pfnActiveTexture			( NULL)
	,	m_pfnClientActiveTexture	( NULL)
	,	m_pfnGetTexImage			( NULL)
	,	m_pfnTexEnvi				( NULL)
	,	m_pfnTexEnviv				( NULL)
	,	m_pfnTexEnvf				( NULL)
	,	m_pfnTexEnvfv				( NULL)
	,	m_pfnBlendFunc				( NULL)
	,	m_pfnAlphaFunc				( NULL)
	,	m_pfnEnableClientState		( NULL)
	,	m_pfnDisableClientState		( NULL)
	,	m_pfnGenBuffers				( NULL)
	,	m_pfnBindBuffer				( NULL)
	,	m_pfnDeleteBuffers			( NULL)
	,	m_pfnIsBuffer				( NULL)
	,	m_pfnBufferData				( NULL)
	,	m_pfnBufferSubData			( NULL)
	,	m_pfnGetBufferParameteriv	( NULL)
	,	m_pfnMapBuffer				( NULL)
	,	m_pfnUnmapBuffer			( NULL)
	,	m_pfnVertexPointer			( NULL)
	,	m_pfnNormalPointer			( NULL)
	,	m_pfnTexCoordPointer		( NULL)
	,	m_pfnGenFramebuffers		( NULL)
	,	m_pfnDeleteFramebuffers		( NULL)
	,	m_pfnBindFramebuffer		( NULL)
	,	m_pfnFramebufferTexture2D	( NULL)
	,	m_pfnCheckFramebufferStatus	( NULL)
	,	m_pfnGetUniformLocation		( NULL)
	,	m_pfnUniform1i				( NULL)
	,	m_pfnUniform1f				( NULL)
	,	m_pfnCreateShader			( NULL)
	,	m_pfnDeleteShader			( NULL)
	,	m_pfnAttachShader			( NULL)
	,	m_pfnDetachShader			( NULL)
	,	m_pfnCompileShader			( NULL)
	,	m_pfnGetShaderiv			( NULL)
	,	m_pfnGetShaderInfoLog		( NULL)
	,	m_pfnShaderSource			( NULL)
	,	m_pfnCreateProgram			( NULL)
	,	m_pfnDeleteProgram			( NULL)
	,	m_pfnLinkProgram			( NULL)
	,	m_pfnUseProgram				( NULL)
	,	m_pfnGetProgramiv			( NULL)
	,	m_pfnGetProgramInfoLog		( NULL)
	,	m_pfnGetAttribLocation		( NULL)
	,	m_pfnIsProgram				( NULL)
{
}

OpenGl :: ~OpenGl()
{
}

void OpenGl :: Initialise()
{
	if (m_instance == NULL)
	{
		m_instance = new OpenGl();

		glewInit();

		if (glGenerateMipmap != NULL)
		{
			m_instance->m_pfnGenerateMipmap			= glGenerateMipmap			;
		}
		else
		{
			m_instance->m_pfnGenerateMipmap			= glGenerateMipmapEXT		;
		}

		m_instance->m_pfnGetError					= & glGetError				;
		m_instance->m_pfnActiveTexture				= glActiveTexture			;
		m_instance->m_pfnClientActiveTexture		= glClientActiveTexture		;
		m_instance->m_pfnGenBuffers					= glGenBuffers				;
		m_instance->m_pfnBindBuffer					= glBindBuffer				;
		m_instance->m_pfnDeleteBuffers				= glDeleteBuffers			;
		m_instance->m_pfnIsBuffer					= glIsBuffer				;
		m_instance->m_pfnBufferData					= glBufferData				;
		m_instance->m_pfnBufferSubData				= glBufferSubData			;
		m_instance->m_pfnGetBufferParameteriv		= glGetBufferParameteriv	;
		m_instance->m_pfnMapBuffer					= glMapBuffer				;
		m_instance->m_pfnUnmapBuffer				= glUnmapBuffer				;
		m_instance->m_pfnGenFramebuffers			= glGenFramebuffers			;
		m_instance->m_pfnDeleteFramebuffers			= glDeleteFramebuffers		;
		m_instance->m_pfnBindFramebuffer			= glBindFramebuffer			;
		m_instance->m_pfnFramebufferTexture2D		= glFramebufferTexture2D	;
		m_instance->m_pfnCheckFramebufferStatus		= glCheckFramebufferStatus	;
		m_instance->m_pfnGetUniformLocation			= glGetUniformLocation		;
		m_instance->m_pfnUniform1i					= glUniform1i				;
		m_instance->m_pfnUniform1f					= glUniform1f				;
		m_instance->m_pfnCreateShader				= glCreateShader			;
		m_instance->m_pfnDeleteShader				= glDeleteShader			;
		m_instance->m_pfnAttachShader				= glAttachShader			;
		m_instance->m_pfnDetachShader				= glDetachShader			;
		m_instance->m_pfnCompileShader				= glCompileShader			;
		m_instance->m_pfnGetShaderiv				= glGetShaderiv				;
		m_instance->m_pfnGetShaderInfoLog			= glGetShaderInfoLog		;
		m_instance->m_pfnShaderSource				= glShaderSource			;
		m_instance->m_pfnCreateProgram				= glCreateProgram			;
		m_instance->m_pfnDeleteProgram				= glDeleteProgram			;
		m_instance->m_pfnLinkProgram				= glLinkProgram				;
		m_instance->m_pfnUseProgram					= glUseProgram				;
		m_instance->m_pfnGetProgramiv				= glGetProgramiv			;
		m_instance->m_pfnGetProgramInfoLog			= glGetProgramInfoLog		;
		m_instance->m_pfnGetAttribLocation			= glGetAttribLocation		;
		m_instance->m_pfnIsProgram					= glIsProgram				;
		m_instance->m_pfnClearColor					= & glClearColor			;
		m_instance->m_pfnShadeModel					= & glShadeModel			;
		m_instance->m_pfnViewport					= & glViewport				;
		m_instance->m_pfnMatrixMode					= & glMatrixMode			;
		m_instance->m_pfnLoadIdentity				= & glLoadIdentity			;
		m_instance->m_pfnOrtho						= & glOrtho					;
		m_instance->m_pfnClear						= & glClear					;
		m_instance->m_pfnDrawElements				= & glDrawElements			;
		m_instance->m_pfnEnable						= & glEnable				;
		m_instance->m_pfnDisable					= & glDisable				;
		m_instance->m_pfnGenTextures				= & glGenTextures			;
		m_instance->m_pfnDeleteTextures				= & glDeleteTextures		;
		m_instance->m_pfnBindTexture				= & glBindTexture			;
		m_instance->m_pfnTexSubImage2D				= & glTexSubImage2D			;
		m_instance->m_pfnTexImage2D					= & glTexImage2D			;
		m_instance->m_pfnTexParameteri				= & glTexParameteri			;
		m_instance->m_pfnEnableClientState			= & glEnableClientState		;
		m_instance->m_pfnDisableClientState			= & glDisableClientState	;
		m_instance->m_pfnVertexPointer				= & glVertexPointer			;
		m_instance->m_pfnNormalPointer				= & glNormalPointer			;
		m_instance->m_pfnGetTexImage				= & glGetTexImage			;
		m_instance->m_pfnTexCoordPointer			= & glTexCoordPointer		;
		m_instance->m_pfnTexEnvi					= & glTexEnvi				;
		m_instance->m_pfnTexEnviv					= & glTexEnviv				;
		m_instance->m_pfnTexEnvf					= & glTexEnvf				;
		m_instance->m_pfnTexEnvfv					= & glTexEnvfv				;
		m_instance->m_pfnBlendFunc					= & glBlendFunc				;
		m_instance->m_pfnAlphaFunc					= & glAlphaFunc				;
		m_instance->m_pfnReadBuffer					= & glReadBuffer			;
		m_instance->m_pfnReadPixels					= & glReadPixels			;
	}
}

void OpenGl :: Cleanup()
{
	if (m_instance != NULL)
	{
		m_instance->m_pfnGetError					= NULL;
		m_instance->m_pfnGenerateMipmap				= NULL;
		m_instance->m_pfnActiveTexture				= NULL;
		m_instance->m_pfnClientActiveTexture		= NULL;
		m_instance->m_pfnGenBuffers					= NULL;
		m_instance->m_pfnBindBuffer					= NULL;
		m_instance->m_pfnDeleteBuffers				= NULL;
		m_instance->m_pfnIsBuffer					= NULL;
		m_instance->m_pfnBufferData					= NULL;
		m_instance->m_pfnBufferSubData				= NULL;
		m_instance->m_pfnGetBufferParameteriv		= NULL;
		m_instance->m_pfnMapBuffer					= NULL;
		m_instance->m_pfnUnmapBuffer				= NULL;
		m_instance->m_pfnGenFramebuffers			= NULL;
		m_instance->m_pfnDeleteFramebuffers			= NULL;
		m_instance->m_pfnBindFramebuffer			= NULL;
		m_instance->m_pfnFramebufferTexture2D		= NULL;
		m_instance->m_pfnCheckFramebufferStatus		= NULL;
		m_instance->m_pfnGetUniformLocation			= NULL;
		m_instance->m_pfnUniform1i					= NULL;
		m_instance->m_pfnUniform1f					= NULL;
		m_instance->m_pfnCreateShader				= NULL;
		m_instance->m_pfnDeleteShader				= NULL;
		m_instance->m_pfnAttachShader				= NULL;
		m_instance->m_pfnDetachShader				= NULL;
		m_instance->m_pfnCompileShader				= NULL;
		m_instance->m_pfnGetShaderiv				= NULL;
		m_instance->m_pfnGetShaderInfoLog			= NULL;
		m_instance->m_pfnShaderSource				= NULL;
		m_instance->m_pfnCreateProgram				= NULL;
		m_instance->m_pfnDeleteProgram				= NULL;
		m_instance->m_pfnLinkProgram				= NULL;
		m_instance->m_pfnUseProgram					= NULL;
		m_instance->m_pfnGetProgramiv				= NULL;
		m_instance->m_pfnGetProgramInfoLog			= NULL;
		m_instance->m_pfnGetAttribLocation			= NULL;
		m_instance->m_pfnIsProgram					= NULL;
		m_instance->m_pfnClearColor					= NULL;
		m_instance->m_pfnShadeModel					= NULL;
		m_instance->m_pfnViewport					= NULL;
		m_instance->m_pfnMatrixMode					= NULL;
		m_instance->m_pfnLoadIdentity				= NULL;
		m_instance->m_pfnOrtho						= NULL;
		m_instance->m_pfnClear						= NULL;
		m_instance->m_pfnDrawElements				= NULL;
		m_instance->m_pfnEnable						= NULL;
		m_instance->m_pfnDisable					= NULL;
		m_instance->m_pfnGenTextures				= NULL;
		m_instance->m_pfnDeleteTextures				= NULL;
		m_instance->m_pfnBindTexture				= NULL;
		m_instance->m_pfnTexSubImage2D				= NULL;
		m_instance->m_pfnTexImage2D					= NULL;
		m_instance->m_pfnTexParameteri				= NULL;
		m_instance->m_pfnEnableClientState			= NULL;
		m_instance->m_pfnDisableClientState			= NULL;
		m_instance->m_pfnVertexPointer				= NULL;
		m_instance->m_pfnNormalPointer				= NULL;
		m_instance->m_pfnGetTexImage				= NULL;
		m_instance->m_pfnTexCoordPointer			= NULL;
		m_instance->m_pfnTexEnvi					= NULL;
		m_instance->m_pfnBlendFunc					= NULL;
		m_instance->m_pfnAlphaFunc					= NULL;
		m_instance->m_pfnReadBuffer					= NULL;
		m_instance->m_pfnReadPixels					= NULL;
		delete m_instance;
		m_instance = NULL;
	}
}

GLenum OpenGl :: GetError()
{
	return m_instance->m_pfnGetError();
}

bool OpenGl :: ClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha)
{
	m_instance->m_pfnClearColor( red, green, blue, alpha);
	return glCheckError( "glClearColor");
}

bool OpenGl :: ShadeModel( GLenum mode)
{
	m_instance->m_pfnShadeModel( mode);
	return glCheckError( "glShadeModel");
}

bool OpenGl :: Viewport( GLint x, GLint y, GLsizei width, GLsizei height)
{
	m_instance->m_pfnViewport( x, y, width, height);
	return glCheckError( "glViewport");
}

bool OpenGl :: MatrixMode( GLenum mode)
{
	m_instance->m_pfnMatrixMode( mode);
	return glCheckError( "glMatrixMode");
}

bool OpenGl :: LoadIdentity()
{
	m_instance->m_pfnLoadIdentity();
	return glCheckError( "glLoadIdentity");
}

bool OpenGl :: Ortho( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar)
{
	m_instance->m_pfnOrtho( left, right, bottom, top, zNear, zFar);
	return glCheckError( "glOrtho");
}

bool OpenGl :: Clear( GLbitfield mask)
{
	m_instance->m_pfnClear( mask);
	return glCheckError( "glClear");
}

bool OpenGl :: DrawElements( GLenum mode, GLsizei count, GLenum type, const GLvoid *indices)
{
	m_instance->m_pfnDrawElements( mode, count, type, indices);
	return glCheckError( "glDrawElements");
}

bool OpenGl :: Enable( GLenum mode)
{
	m_instance->m_pfnEnable( mode);
	return glCheckError( "glEnable");
}

bool OpenGl :: Disable( GLenum mode)
{
	m_instance->m_pfnDisable( mode);
	return glCheckError( "glDisable");
}

bool OpenGl :: GenTextures( GLsizei n, GLuint* textures)
{
	m_instance->m_pfnGenTextures( n, textures);
	return glCheckError( "glGenTextures");
}

bool OpenGl :: DeleteTextures( GLsizei n, const GLuint * textures)
{
	m_instance->m_pfnDeleteTextures( n, textures);
	return glCheckError( "glDeleteTextures");
}

bool OpenGl :: GenerateMipmap( GLenum target)
{
	m_instance->m_pfnGenerateMipmap( target);
	return glCheckError( "glGenerateMipmap");
}

bool OpenGl :: ActiveTexture( GLint texture)
{
	m_instance->m_pfnActiveTexture( texture);
	return glCheckError( "glActiveTexture");
}

bool OpenGl :: ClientActiveTexture( GLenum texture)
{
	m_instance->m_pfnClientActiveTexture( texture);
	return glCheckError( "glClientActiveTexture");
}

bool OpenGl :: BindTexture( GLenum mode, GLuint texture)
{
	m_instance->m_pfnBindTexture( mode, texture);
	return glCheckError( "glBindTexture");
}

bool OpenGl :: TexSubImage2D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid * data)
{
	m_instance->m_pfnTexSubImage2D( target, level, xoffset, yoffset, width, height, format, type, data);
	return glCheckError( "glTexSubImage2D");
}

bool OpenGl :: TexImage2D( GLenum target, GLint level, GLint internalFormat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid * data)
{
	m_instance->m_pfnTexImage2D( target, level, internalFormat, width, height, border, format, type, data);
	return glCheckError( "glTexImage2D");
}

bool OpenGl :: TexParameteri( GLenum target, GLenum pname, GLint param)
{
	m_instance->m_pfnTexParameteri( target, pname, param);
	return glCheckError( "glTexParameteri");
}

bool OpenGl :: GetTexImage( GLenum target, GLint level, GLenum format, GLenum type, GLvoid * img)
{
	m_instance->m_pfnGetTexImage( target, level, format, type, img);
	return glCheckError( "glGetTexImage");
}

bool OpenGl :: TexEnvi( GLenum target, GLenum pname, GLint param)
{
	m_instance->m_pfnTexEnvi( target, pname, param);
	return glCheckError( "glTexEnvi");
}

bool OpenGl :: TexEnviv( GLenum target, GLenum pname, const GLint * param)
{
	m_instance->m_pfnTexEnviv( target, pname, param);
	return glCheckError( "glTexEnviv");
}

bool OpenGl :: TexEnvf( GLenum target, GLenum pname, GLfloat param)
{
	m_instance->m_pfnTexEnvf( target, pname, param);
	return glCheckError( "glTexEnvf");
}

bool OpenGl :: TexEnvfv( GLenum target, GLenum pname, const GLfloat * param)
{
	m_instance->m_pfnTexEnvfv( target, pname, param);
	return glCheckError( "glTexEnvfv");
}

bool OpenGl :: BlendFunc( GLenum sfactor, GLenum dfactor)
{
	m_instance->m_pfnBlendFunc( sfactor, dfactor);
	return glCheckError( "glBlendFunc");
}

bool OpenGl :: AlphaFunc( GLenum func, GLclampf ref)
{
	m_instance->m_pfnAlphaFunc( func, ref);
	return glCheckError( "glAlphaFunc");
}

bool OpenGl :: ReadBuffer( GLenum mode)
{
	m_instance->m_pfnReadBuffer( mode);
	return glCheckError( "glReadBuffer");
}

bool OpenGl :: ReadPixels( GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid * pixels)
{
	m_instance->m_pfnReadPixels( x, y, width, height, format, type, pixels);
	return glCheckError( "glReadPixels");
}

bool OpenGl :: EnableClientState( GLenum array)
{
	m_instance->m_pfnEnableClientState( array);
	return glCheckError( "glEnableClientState");
}

bool OpenGl :: DisableClientState( GLenum array)
{
	m_instance->m_pfnDisableClientState( array);
	return glCheckError( "glDisableClientState");
}

bool OpenGl :: GenBuffers( GLsizei n, GLuint* buffers)
{
	m_instance->m_pfnGenBuffers( n, buffers);
	return glCheckError( "glGenBuffers");
}

bool OpenGl :: BindBuffer( GLenum target, GLuint buffer)
{
	m_instance->m_pfnBindBuffer( target, buffer);
	return glCheckError( "glBindBuffer");
}

bool OpenGl :: DeleteBuffers( GLsizei n, const GLuint* buffers)
{
	m_instance->m_pfnDeleteBuffers( n, buffers);
	return glCheckError( "glDeleteBuffers");
}

bool OpenGl :: IsBuffer( GLuint buffer)
{
	return m_instance->m_pfnIsBuffer( buffer) && glCheckError( "glIsBuffer");
}

bool OpenGl :: BufferData( GLenum target, GLsizeiptr size, const GLvoid* data, GLenum usage)
{
	m_instance->m_pfnBufferData( target, size, data, usage);
	return glCheckError( "glBufferData");
}

bool OpenGl :: BufferSubData( GLenum target, GLintptrARB offset, GLsizeiptrARB size, const GLvoid * data)
{
	m_instance->m_pfnBufferSubData( target, offset, size, data);
	return glCheckError( "glBufferSubData");
}

bool OpenGl :: GetBufferParameteriv( GLenum target, GLenum pname, GLint * params)
{
	m_instance->m_pfnGetBufferParameteriv( target, pname, params);
	return glCheckError( "glGetBufferParameteriv");
}

GLvoid * OpenGl :: MapBuffer( GLenum target, GLenum access)
{
	GLvoid * l_pReturn = m_instance->m_pfnMapBuffer( target, access);

	if ( ! glCheckError( "glUnmapBuffer"))
	{
		l_pReturn = NULL;
	}

	return l_pReturn;
}

bool OpenGl :: UnmapBuffer( GLenum target)
{
	m_instance->m_pfnUnmapBuffer( target);
	return glCheckError( "glUnmapBuffer");
}

bool OpenGl :: VertexPointer( GLint size, GLenum type, GLsizei stride, const GLvoid * pointer)
{
	m_instance->m_pfnVertexPointer( size, type, stride, pointer);
	return glCheckError( "glVertexPointer");
}

bool OpenGl :: NormalPointer( GLenum type, GLsizei stride, const GLvoid * pointer)
{
	m_instance->m_pfnNormalPointer( type, stride, pointer);
	return glCheckError( "glNormalPointer");
}

bool OpenGl :: TexCoordPointer( GLint size, GLenum type, GLsizei stride, const GLvoid * pointer)
{
	m_instance->m_pfnTexCoordPointer( size, type, stride, pointer);
	return glCheckError( "glTexCoordPointer");
}

bool OpenGl :: GenFramebuffers( GLsizei n, GLuint* framebuffers)
{
	m_instance->m_pfnGenFramebuffers( n, framebuffers);
	return glCheckError( "glGenFramebuffers");
}

bool OpenGl :: DeleteFramebuffers( GLsizei n, const GLuint* framebuffers)
{
	m_instance->m_pfnDeleteFramebuffers( n, framebuffers);
	return glCheckError( "glDeleteFramebuffers");
}

bool OpenGl :: BindFramebuffer( GLenum target, GLuint framebuffer)
{
	m_instance->m_pfnBindFramebuffer( target, framebuffer);
	return glCheckError( "glBindFramebuffer");
}

bool OpenGl :: FramebufferTexture2D( GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level)
{
	m_instance->m_pfnFramebufferTexture2D( target, attachment, textarget, texture, level);
	return glCheckError( "glFramebufferTexture2D");
}

GLenum OpenGl :: CheckFramebufferStatus( GLenum target)
{
	GLenum l_eReturn = m_instance->m_pfnCheckFramebufferStatus( target);

	if ( ! glCheckError( "glCheckFramebufferStatus"))
	{
		l_eReturn = GL_INVALID_INDEX;
	}

	return l_eReturn;
}

GLint OpenGl :: GetUniformLocation( GLuint program, const GLchar* name)
{
	GLint l_iReturn = m_instance->m_pfnGetUniformLocation( program, name);

	if ( ! glCheckError( "glGetUniformLocation"))
	{
		l_iReturn = GL_INVALID_INDEX;
	}

	return l_iReturn;
}

bool OpenGl :: Uniform1i( GLint location, GLint v0)
{
	m_instance->m_pfnUniform1i( location, v0);
	return glCheckError( "glUniform1i");
}

bool OpenGl :: Uniform1f( GLint location, GLfloat v0)
{
	m_instance->m_pfnUniform1f( location, v0);
	return glCheckError( "glUniform1f");
}

GLuint OpenGl :: CreateShader( GLenum type)
{
	GLuint l_uiReturn = m_instance->m_pfnCreateShader( type);

	if ( ! glCheckError( "glCreateShader"))
	{
		l_uiReturn = GL_INVALID_INDEX;
	}

	return l_uiReturn;
}

bool OpenGl :: DeleteShader( GLuint shader)
{
	m_instance->m_pfnDeleteShader( shader);
	return glCheckError( "glDeleteShader");
}

bool OpenGl :: AttachShader( GLuint program, GLuint shader)
{
	m_instance->m_pfnAttachShader( program, shader);
	return glCheckError( "glAttachShader");
}

bool OpenGl :: DetachShader( GLuint program, GLuint shader)
{
	m_instance->m_pfnDetachShader( program, shader);
	return glCheckError( "glDetachShader");
}

bool OpenGl :: CompileShader( GLuint shader)
{
	m_instance->m_pfnCompileShader( shader);
	return glCheckError( "glCompileShader");
}

bool OpenGl :: GetShaderiv( GLuint shader, GLenum pname, GLint* param)
{
	m_instance->m_pfnGetShaderiv( shader, pname, param);
	return glCheckError( "glGetShaderiv");
}

bool OpenGl :: GetShaderInfoLog( GLuint shader, GLsizei bufSize, GLsizei* length, GLchar* infoLog)
{
	m_instance->m_pfnGetShaderInfoLog( shader, bufSize, length, infoLog);
	return glCheckError( "glGetShaderInfoLog");
}

bool OpenGl :: ShaderSource( GLuint shader, GLsizei count, const GLchar** strings, const GLint* lengths)
{
	m_instance->m_pfnShaderSource( shader, count, strings, lengths);
	return glCheckError( "glShaderSource");
}

GLuint OpenGl :: CreateProgram()
{
	GLuint l_uiReturn = m_instance->m_pfnCreateProgram();

	if ( ! glCheckError( "glCreateProgram"))
	{
		l_uiReturn = GL_INVALID_INDEX;
	}

	return l_uiReturn;
}

bool OpenGl :: DeleteProgram( GLuint program)
{
	m_instance->m_pfnDeleteProgram( program);
	return glCheckError( "glDeleteProgram");
}

bool OpenGl :: LinkProgram( GLuint program)
{
	m_instance->m_pfnLinkProgram( program);
	return glCheckError( "glLinkProgram");
}

bool OpenGl :: UseProgram( GLuint program)
{
	m_instance->m_pfnUseProgram( program);
	return glCheckError( "glUseProgram");
}

bool OpenGl :: GetProgramiv( GLuint program, GLenum pname, GLint* param)
{
	m_instance->m_pfnGetProgramiv( program, pname, param);
	return glCheckError( "glGetProgramiv");
}

bool OpenGl :: GetProgramInfoLog( GLuint program, GLsizei bufSize, GLsizei* length, GLchar* infoLog)
{
	m_instance->m_pfnGetProgramInfoLog( program, bufSize, length, infoLog);
	return glCheckError( "glGetProgramInfoLog");
}

GLint OpenGl :: GetAttribLocation( GLuint program, const GLchar* name)
{
	GLint l_iReturn = m_instance->m_pfnGetAttribLocation( program, name);

	if ( ! glCheckError( "glGetAttribLocation"))
	{
		l_iReturn = GL_INVALID_INDEX;
	}

	return l_iReturn;
}

bool OpenGl :: IsProgram( GLuint program)
{
	m_instance->m_pfnIsProgram( program);
	return glCheckError( "glIsProgram");
}

//*************************************************************************************************

GlBuffer :: GlBuffer( GLenum p_target, GLenum p_mode)
	:	m_uiGlIndex( GL_INVALID_INDEX)
	,	m_target( p_target)
	,	m_mode( p_mode)
{
}

GlBuffer :: ~GlBuffer()
{
}

void GlBuffer :: Create()
{
	OpenGl::GenBuffers( 1, & m_uiGlIndex);
}

void GlBuffer :: Cleanup()
{
	if (m_uiGlIndex != GL_INVALID_INDEX)
	{
		OpenGl::DeleteBuffers( 1, & m_uiGlIndex);
		m_uiGlIndex = GL_INVALID_INDEX;
	}
}

bool GlBuffer :: Bind()
{
	return OpenGl::BindBuffer( m_target, m_uiGlIndex);
}

bool GlBuffer :: Unbind()
{
	return OpenGl::BindBuffer( m_target, 0);
}

void GlBuffer :: Data( void * p_pBuffer, size_t p_uiSize)
{
	OpenGl::BufferData( m_target, p_uiSize, p_pBuffer, m_mode);
}

void * GlBuffer :: Lock( GLenum p_access)
{
	void * l_pReturn = NULL;

	if (m_uiGlIndex != GL_INVALID_INDEX)
	{
		l_pReturn = OpenGl::MapBuffer( m_target, p_access);
	}

	return l_pReturn;
}

bool GlBuffer :: Unlock()
{
	bool l_bReturn = m_uiGlIndex != GL_INVALID_INDEX;
	l_bReturn &= OpenGl::UnmapBuffer( m_target);
	return l_bReturn;
}

//*************************************************************************************************

GlVertexBuffer :: GlVertexBuffer()
	:	GlBuffer( GL_ARRAY_BUFFER, GL_STATIC_DRAW)
{
	m_vertexDeclarationElements = new BufferElementDeclaration[4];

	ProceduralTextures::Initialise( m_vertexDeclarationElements[0], 0, eUsagePosition,		eType3Floats, 0 );
	ProceduralTextures::Initialise( m_vertexDeclarationElements[1], 0, eUsageNormal,		eType3Floats, 0 );
	ProceduralTextures::Initialise( m_vertexDeclarationElements[2], 0, eUsageTangent,		eType3Floats, 0 );
	ProceduralTextures::Initialise( m_vertexDeclarationElements[3], 0, eUsageTexCoords0,	eType2Floats, 0 );

	m_vertexDeclaration = new BufferDeclaration( m_vertexDeclarationElements, 4 );
	Vertex l_vtxTmp( m_vertexDeclaration );

	m_arrayVertex.push_back( l_vtxTmp );
	m_arrayVertex.push_back( l_vtxTmp );
	m_arrayVertex.push_back( l_vtxTmp );
	m_arrayVertex.push_back( l_vtxTmp );

	m_arrayVertex[0].LinkCoords( & m_pBuffer[0]);
	m_arrayVertex[1].LinkCoords( & m_pBuffer[11]);
	m_arrayVertex[2].LinkCoords( & m_pBuffer[22]);
	m_arrayVertex[3].LinkCoords( & m_pBuffer[33]);

	m_arrayVertex[0].SetCoords( 0.0f, 0.0f, 0.0f);
	m_arrayVertex[1].SetCoords( 0.0f, 1.0f, 0.0f);
	m_arrayVertex[2].SetCoords( 1.0f, 1.0f, 0.0f);
	m_arrayVertex[3].SetCoords( 1.0f, 0.0f, 0.0f);

	m_arrayVertex[0].SetNormal( 0.0f, 0.0f, -1.0f);
	m_arrayVertex[1].SetNormal( 0.0f, 0.0f, -1.0f);
	m_arrayVertex[2].SetNormal( 0.0f, 0.0f, -1.0f);
	m_arrayVertex[3].SetNormal( 0.0f, 0.0f, -1.0f);

	m_arrayVertex[0].SetTexCoord( 0.0f, 0.0f);
	m_arrayVertex[1].SetTexCoord( 0.0f, 1.0f);
	m_arrayVertex[2].SetTexCoord( 1.0f, 1.0f);
	m_arrayVertex[3].SetTexCoord( 1.0f, 0.0f);

	const Point3f & l_v0 = m_arrayVertex[0].GetCoords();
	const Point3f & l_v1 = m_arrayVertex[1].GetCoords();
	const Point3f & l_v2 = m_arrayVertex[2].GetCoords();

	const Point2f & l_w0 = m_arrayVertex[0].GetTexCoord();
	const Point2f & l_w1 = m_arrayVertex[1].GetTexCoord();
	const Point2f & l_w2 = m_arrayVertex[2].GetTexCoord();

	Point3f l_vEdge1( l_v1 - l_v0);
	Point3f l_vEdge2( l_v2 - l_v0);
	Point2f l_vEdge1uv( l_w1 - l_w0);
	Point2f l_vEdge2uv( l_w2 - l_w0);

	float l_fCoeff = l_vEdge1uv[1] * l_vEdge2uv[0] - l_vEdge1uv[0] * l_vEdge2uv[1];

	if (l_fCoeff != 0.0f)
	{
		float l_fMult = 1.0f / l_fCoeff;
		m_arrayVertex[0].SetTangent( (l_vEdge1 * -l_vEdge2uv[1] + l_vEdge2 * l_vEdge1uv[1]) * l_fMult);
		m_arrayVertex[1].SetTangent( (l_vEdge1 * -l_vEdge2uv[1] + l_vEdge2 * l_vEdge1uv[1]) * l_fMult);
		m_arrayVertex[2].SetTangent( (l_vEdge1 * -l_vEdge2uv[1] + l_vEdge2 * l_vEdge1uv[1]) * l_fMult);
		m_arrayVertex[3].SetTangent( (l_vEdge1 * -l_vEdge2uv[1] + l_vEdge2 * l_vEdge1uv[1]) * l_fMult);
	}
}

GlVertexBuffer :: ~GlVertexBuffer()
{
	delete m_vertexDeclaration;
	delete [] m_vertexDeclarationElements;
}

void GlVertexBuffer :: Initialise()
{
	Create();

	if (Bind())
	{
		Data( m_pBuffer, sizeof( m_pBuffer));
		Unbind();
	}
}

bool GlVertexBuffer :: Activate()
{
	bool l_bReturn = Bind();
	l_bReturn &= OpenGl::EnableClientState( GL_VERTEX_ARRAY);
	l_bReturn &= OpenGl::VertexPointer( 3, GL_FLOAT, m_vertexDeclaration->GetStride(), GL_BUFFER_OFFSET( 0));
	l_bReturn &= OpenGl::EnableClientState( GL_NORMAL_ARRAY);
	l_bReturn &= OpenGl::NormalPointer( GL_FLOAT, m_vertexDeclaration->GetStride(), GL_BUFFER_OFFSET( 12));
	l_bReturn &= OpenGl::ClientActiveTexture( GL_TEXTURE0);
	l_bReturn &= OpenGl::EnableClientState( GL_TEXTURE_COORD_ARRAY);
	l_bReturn &= OpenGl::TexCoordPointer( 2, GL_FLOAT, m_vertexDeclaration->GetStride(), GL_BUFFER_OFFSET( 36));
	l_bReturn &= OpenGl::ClientActiveTexture( GL_TEXTURE1);
	l_bReturn &= OpenGl::EnableClientState( GL_TEXTURE_COORD_ARRAY);
	l_bReturn &= OpenGl::TexCoordPointer( 2, GL_FLOAT, m_vertexDeclaration->GetStride(), GL_BUFFER_OFFSET( 36));
	l_bReturn &= OpenGl::ClientActiveTexture( GL_TEXTURE2);
	l_bReturn &= OpenGl::EnableClientState( GL_TEXTURE_COORD_ARRAY);
	l_bReturn &= OpenGl::TexCoordPointer( 2, GL_FLOAT, m_vertexDeclaration->GetStride(), GL_BUFFER_OFFSET( 36));
	l_bReturn &= OpenGl::ClientActiveTexture( GL_TEXTURE3);
	l_bReturn &= OpenGl::EnableClientState( GL_TEXTURE_COORD_ARRAY);
	l_bReturn &= OpenGl::TexCoordPointer( 2, GL_FLOAT, m_vertexDeclaration->GetStride(), GL_BUFFER_OFFSET( 36));
	return l_bReturn;
}

void GlVertexBuffer :: Deactivate()
{
	OpenGl::DisableClientState( GL_VERTEX_ARRAY);
	OpenGl::DisableClientState( GL_NORMAL_ARRAY);
	OpenGl::ClientActiveTexture( GL_TEXTURE0);
	OpenGl::DisableClientState( GL_TEXTURE_COORD_ARRAY);
	OpenGl::ClientActiveTexture( GL_TEXTURE1);
	OpenGl::DisableClientState( GL_TEXTURE_COORD_ARRAY);
	OpenGl::ClientActiveTexture( GL_TEXTURE2);
	OpenGl::DisableClientState( GL_TEXTURE_COORD_ARRAY);
	OpenGl::ClientActiveTexture( GL_TEXTURE3);
	OpenGl::DisableClientState( GL_TEXTURE_COORD_ARRAY);
	Unbind();
}

//*************************************************************************************************

GlIndexBuffer :: GlIndexBuffer()
	:	GlBuffer( GL_ELEMENT_ARRAY_BUFFER, GL_STATIC_DRAW)
{
	m_pBuffer[0] = 0;
	m_pBuffer[1] = 1;
	m_pBuffer[2] = 2;
	m_pBuffer[3] = 0;
	m_pBuffer[4] = 2;
	m_pBuffer[5] = 3;
}

GlIndexBuffer :: ~GlIndexBuffer()
{
}

void GlIndexBuffer :: Initialise()
{
	Create();

	if (Bind())
	{
		Data( m_pBuffer, sizeof( m_pBuffer));
		Unbind();
	}
}

bool GlIndexBuffer :: Activate()
{
	return Bind();
}

void GlIndexBuffer :: Deactivate()
{
	Unbind();
}

//*************************************************************************************************

GlPixelBuffer :: GlPixelBuffer( PixelBuffer * p_pPixels, GLenum p_packMode, GLenum p_drawMode)
	:	GlBuffer( p_packMode, p_drawMode)
	,	m_pPixels	( p_pPixels)
	,	m_packMode	( p_packMode)
	,	m_drawMode	( p_drawMode)
{
}

GlPixelBuffer :: ~GlPixelBuffer()
{
}

bool GlPixelBuffer :: Activate()
{
	return Bind();
}

void GlPixelBuffer :: Deactivate()
{
	Unbind();
}

//*************************************************************************************************

GlPackPixelBuffer :: GlPackPixelBuffer( PixelBuffer * p_pPixels)
	:	GlPixelBuffer( p_pPixels, GL_PIXEL_PACK_BUFFER, GL_STREAM_READ)
{
}

GlPackPixelBuffer :: ~GlPackPixelBuffer()
{
}

void GlPackPixelBuffer :: Initialise()
{
	Create();

	if (Bind())
	{
		if (m_pPixels != NULL)
		{
			Data( m_pPixels->ptr(), m_pPixels->size());
		}

		Unbind();
	}
}

//*************************************************************************************************

GlUnpackPixelBuffer :: GlUnpackPixelBuffer( PixelBuffer * p_pPixels)
	:	GlPixelBuffer( p_pPixels, GL_PIXEL_UNPACK_BUFFER, GL_STREAM_DRAW)
{
}

GlUnpackPixelBuffer :: ~GlUnpackPixelBuffer()
{
}

void GlUnpackPixelBuffer :: Initialise()
{
	Create();

	if (Bind())
	{
		if (m_pPixels != NULL)
		{
			Data( NULL, m_pPixels->size());
		}

		Unbind();
	}
}

//*************************************************************************************************

GlTexture :: GlTexture()
	:	m_uiGlTexture	( GL_INVALID_INDEX)
	,	m_bInitialised	( false)
	,	m_pPixels		( NULL)
	,	m_bOwnsPixels	( true)
	,	m_pxColour		( 1.0f, 1.0f, 1.0f, 1.0f)
	,	m_iCurrentPbo	( 0 )
{
	m_pIoBuffer = NULL;
	m_pIoBuffers[0] = NULL;
	m_pIoBuffers[1] = NULL;
}

GlTexture :: ~GlTexture()
{
	DoDeleteBuffers();
}

void GlTexture :: Initialise()
{
	OpenGl::GenTextures( 1, & m_uiGlTexture);
}

void GlTexture :: Initialise( PixelBuffer * p_pBuffer)
{
	if ( ! m_bInitialised)
	{
		DoCleanupPbos();
		DoDeleteBuffers();
		m_pPixels = p_pBuffer;
		DoInitialisePbos();
		m_bOwnsPixels = false;
		DoInitGl();
		m_bInitialised = true;
	}
}

void GlTexture :: Initialise( const wxSize & p_size)
{
	if ( ! m_bInitialised)
	{
		DoCleanupPbos();
		DoDeleteBuffers();
		m_size = p_size;
		m_pPixels = new PixelBuffer( Point< size_t, 2 >( p_size.x, p_size.y ) );
		DoInitialisePbos();
		m_bOwnsPixels = true;
		DoInitGl();
		m_bInitialised = true;
	}
}

void GlTexture :: Cleanup()
{
	if( m_uiGlTexture != GL_INVALID_INDEX )
	{
		OpenGl::DeleteTextures( 1, & m_uiGlTexture );
		m_uiGlTexture = GL_INVALID_INDEX;
	}

	DoCleanupPbos();
	DoDeleteBuffers();
}

bool GlTexture :: Activate( GLenum p_texTarget, GLenum p_mode, BlendMode * p_pBlendRgb, BlendMode * p_pBlendAlpha, bool p_bUpload)
{
	bool l_bReturn = OpenGl::ActiveTexture( p_texTarget);
	l_bReturn &= OpenGl::Enable( GL_TEXTURE_2D);
	l_bReturn &= OpenGl::BindTexture( GL_TEXTURE_2D, m_uiGlTexture);

	if (p_bUpload)
	{
		UploadGlImage();
	}

	l_bReturn &= OpenGl::TexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, & m_pxColour.r);
	l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, p_mode);

	if (p_mode == GL_COMBINE)
	{
		if (p_pBlendRgb != NULL && p_pBlendRgb->function != GL_NONE)
		{
			// RGB Combine
			l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_COMBINE_RGB, p_pBlendRgb->function);

			if (p_pBlendRgb->source1 != GL_NONE && p_pBlendRgb->operand1 != GL_NONE)
			{
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_SOURCE0_RGB, p_pBlendRgb->source1);
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_OPERAND0_RGB, p_pBlendRgb->operand1);
			}

			if (p_pBlendRgb->function != GL_REPLACE && p_pBlendRgb->source2 != GL_NONE && p_pBlendRgb->operand2 != GL_NONE)
			{
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_SOURCE1_RGB, p_pBlendRgb->source2);
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_OPERAND1_RGB, p_pBlendRgb->operand2);
			}

			if (p_pBlendRgb->function == GL_INTERPOLATE && p_pBlendRgb->source3 != GL_NONE && p_pBlendRgb->operand3 != GL_NONE)
			{
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_SOURCE2_RGB, p_pBlendRgb->source3);
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_OPERAND2_RGB, p_pBlendRgb->operand3);
			}
		}

		if (p_pBlendAlpha != NULL && p_pBlendAlpha->function != GL_NONE)
		{
			// Alpha Combine
			l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_COMBINE_ALPHA, p_pBlendAlpha->function);

			if (p_pBlendAlpha->source1 != GL_NONE && p_pBlendAlpha->operand1 != GL_NONE)
			{
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_SOURCE0_ALPHA, p_pBlendAlpha->source1);
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, p_pBlendAlpha->operand1);
			}

			if (p_pBlendAlpha->function != GL_REPLACE && p_pBlendAlpha->source2 != GL_NONE && p_pBlendAlpha->operand2 != GL_NONE)
			{
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_SOURCE1_ALPHA, p_pBlendAlpha->source2);
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_OPERAND1_ALPHA, p_pBlendAlpha->operand2);
			}

			if (p_pBlendAlpha->function == GL_INTERPOLATE && p_pBlendAlpha->source3 != GL_NONE && p_pBlendAlpha->operand3 != GL_NONE)
			{
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_SOURCE2_ALPHA, p_pBlendAlpha->source3);
				l_bReturn &= OpenGl::TexEnvi( GL_TEXTURE_ENV, GL_OPERAND2_ALPHA, p_pBlendAlpha->operand3);
			}
		}
	}

	return l_bReturn;
}

void GlTexture :: Deactivate( GLenum p_texTarget)
{
	bool l_bReturn = OpenGl::ActiveTexture( p_texTarget);
	l_bReturn &= OpenGl::Enable( GL_TEXTURE_2D);
	l_bReturn &= OpenGl::BindTexture( GL_TEXTURE_2D, 0);
}

void GlTexture :: UploadGlImage()
{
	GlPixelBuffer * pBufferIn = m_pIoBuffers[m_iCurrentPbo];
	m_iCurrentPbo = 1 - m_iCurrentPbo;
	GlPixelBuffer * pBufferOut = m_pIoBuffers[m_iCurrentPbo];

	if( pBufferIn && pBufferOut )
	{
		if( pBufferOut->Activate() )
		{
			OpenGl::TexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, m_size.x, m_size.y, GL_RGBA, GL_UNSIGNED_BYTE, 0);

			if( pBufferIn->Activate() )
			{
				void * pData = pBufferIn->Lock( GL_WRITE_ONLY );
				memcpy( pData, m_pPixels->const_ptr(), m_pPixels->size());
				pBufferIn->Unlock();
			}

			pBufferOut->Deactivate();
		}
	}
}

void GlTexture :: DownloadGlImage()
{
	if (m_pIoBuffer != NULL)
	{
		m_pIoBuffer->Activate();
		m_pIoBuffer->Data( NULL, m_pPixels->size() );
		void * l_pData = m_pIoBuffer->Lock( GL_READ_ONLY );
		memcpy( m_pPixels->ptr(), l_pData, m_pPixels->size());
		m_pIoBuffer->Unlock();
	}
}
/*
void * GlTexture :: LockGlImage( GLenum p_mode)
{
	void * l_pReturn = NULL;
	GlPixelBuffer * l_pPbo = m_pPixelBuffers[p_mode];

	if (l_pPbo != NULL)
	{
		l_pPbo->Activate();
		l_pPbo->Data( NULL, m_pPixels->size());
		l_pReturn = l_pPbo->Lock( p_mode);
	}

	return l_pReturn;
}

void GlTexture :: UnlockGlImage( GLenum p_mode)
{
	GlPixelBuffer * l_pPbo = m_pPixelBuffers[p_mode];

	if (l_pPbo != NULL)
	{
		l_pPbo->Unlock();

		if (p_mode == GL_WRITE_ONLY)
		{
			OpenGl::TexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, m_size.x, m_size.y, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		}

		l_pPbo->Deactivate();
	}
}
*/
void GlTexture :: Resize( const wxSize & p_size)
{
	m_size = p_size;
	m_bInitialised = false;
}

void GlTexture :: DoInitGl()
{
	bool l_bResult = OpenGl::BindTexture( GL_TEXTURE_2D, m_uiGlTexture);
	l_bResult &= OpenGl::TexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	l_bResult &= OpenGl::TexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	l_bResult &= OpenGl::TexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	l_bResult &= OpenGl::TexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	l_bResult &= OpenGl::TexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	l_bResult &= OpenGl::TexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, m_size.x, m_size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, m_pPixels->const_ptr());
	l_bResult &= OpenGl::BindTexture( GL_TEXTURE_2D, 0);
}

void GlTexture :: DoDeleteBuffers()
{
	if (m_bOwnsPixels)
	{
		delete m_pPixels;
	}

	if (m_pIoBuffers[0] != NULL)
	{
		delete m_pIoBuffers[0];
		m_pIoBuffers[0] = NULL;
	}

	if (m_pIoBuffers[1] != NULL)
	{
		delete m_pIoBuffers[1];
		m_pIoBuffers[1] = NULL;
	}

	if (m_pIoBuffer != NULL)
	{
		delete m_pIoBuffer;
		m_pIoBuffer = NULL;
	}

	m_pPixels = NULL;
	m_bOwnsPixels = true;
}

void GlTexture :: DoCleanupPbos()
{
	if (m_pIoBuffers[0] != NULL)
	{
		m_pIoBuffers[0]->Cleanup();
	}

	if (m_pIoBuffers[1] != NULL)
	{
		m_pIoBuffers[1]->Cleanup();
	}

	if (m_pIoBuffer != NULL)
	{
		m_pIoBuffer->Cleanup();
	}
}

void GlTexture :: DoInitialisePbos()
{
	m_pIoBuffers[0] = new GlUnpackPixelBuffer( m_pPixels);
	m_pIoBuffers[0]->Initialise();
	m_pIoBuffers[1] = new GlUnpackPixelBuffer( m_pPixels);
	m_pIoBuffers[1]->Initialise();
	m_pIoBuffer = new GlUnpackPixelBuffer( m_pPixels);
	m_pIoBuffer->Initialise();
}

//*************************************************************************************************

GlFrameBuffer :: GlFrameBuffer()
	:	m_uiGlBuffer	( GL_INVALID_INDEX)
	,	m_bInitialised	( false)
	,	m_pPixelBuffer	( new GlPackPixelBuffer( NULL))
{
}

GlFrameBuffer :: ~GlFrameBuffer()
{
	delete m_pPixelBuffer;
}

void GlFrameBuffer :: Initialise()
{
	OpenGl::GenFramebuffers( 1, & m_uiGlBuffer);
	m_pPixelBuffer->Initialise();
}

void GlFrameBuffer :: Initialise( ProceduralGenerator * p_pGenerator)
{
	if ( ! m_bInitialised && m_mapAttachments.size() > 0)
	{
		m_pGenerator = p_pGenerator;
		OpenGl::BindFramebuffer( GL_FRAMEBUFFER, m_uiGlBuffer);

		for (std::map<GLenum, GlTexture *>::iterator l_it = m_mapAttachments.begin() ; l_it != m_mapAttachments.end() ; ++l_it)
		{
			OpenGl::FramebufferTexture2D( GL_FRAMEBUFFER, l_it->first, GL_TEXTURE_2D, l_it->second->GetGlIndex(), 0);
		}

		GLenum l_eStatus = OpenGl::CheckFramebufferStatus( GL_FRAMEBUFFER);
		OpenGl::BindFramebuffer( GL_FRAMEBUFFER, 0);

		m_bInitialised = true;
	}
}

void GlFrameBuffer :: Cleanup()
{
	if (m_uiGlBuffer != GL_INVALID_INDEX)
	{
		OpenGl::DeleteFramebuffers( 1, & m_uiGlBuffer);
		m_uiGlBuffer = GL_INVALID_INDEX;
	}

	m_pPixelBuffer->Cleanup();
}

bool GlFrameBuffer :: Activate()
{
	return OpenGl::BindFramebuffer( GL_FRAMEBUFFER, m_uiGlBuffer);
}

void GlFrameBuffer :: Deactivate()
{
	OpenGl::BindFramebuffer( GL_FRAMEBUFFER, 0);
}

bool GlFrameBuffer :: ActivateTexture( GLenum p_attachment)
{
	std::map<GLenum, GlTexture *>::iterator l_it = m_mapAttachments.find( p_attachment);
	bool l_bReturn = l_it != m_mapAttachments.end();

	if (l_bReturn)
	{
		l_bReturn = l_it->second->Activate();
	}

	return l_bReturn;
}

void GlFrameBuffer :: DeactivateTexture( GLenum p_attachment)
{
	std::map<GLenum, GlTexture *>::iterator l_it = m_mapAttachments.find( p_attachment);

	if (l_it != m_mapAttachments.end())
	{
		l_it->second->Deactivate();
	}
}

void GlFrameBuffer :: BindTexture( GLenum p_attachment, GlTexture * p_pTexture)
{
	m_mapAttachments[p_attachment] = p_pTexture;
}

GlTexture * GlFrameBuffer :: GetTexture( GLenum p_attachment)
{
	GlTexture * l_pReturn = NULL;
	std::map<GLenum, GlTexture *>::iterator l_it = m_mapAttachments.find( p_attachment);

	if (l_it != m_mapAttachments.end())
	{
		l_pReturn = l_it->second;
	}

	return l_pReturn;
}
/*
void GlFrameBuffer :: UploadTexture( GLenum p_attachment, const PixelBuffer & p_buffer)
{
	std::map<GLenum, GlTexture *>::iterator l_it = m_mapAttachments.find( p_attachment);

	if (l_it != m_mapAttachments.end())
	{
		void * l_pData = l_it->second->LockGlImage( GL_WRITE_ONLY);
		memcpy( l_pData, p_buffer.const_ptr(), p_buffer.size());
		l_it->second->UnlockGlImage( true);
	}
}

void GlFrameBuffer :: DownloadTexture( GLenum p_attachment, PixelBuffer & p_buffer)
{
	std::map<GLenum, GlTexture *>::iterator l_it = m_mapAttachments.find( p_attachment);

	if (l_it != m_mapAttachments.end())
	{
		void * l_pData = l_it->second->LockGlImage( GL_READ_ONLY);
		memcpy( p_buffer.ptr(), l_pData, p_buffer.size());
		l_it->second->UnlockGlImage( false);
	}
}
*/
bool GlFrameBuffer :: DownloadBuffer( GLenum p_attachment, PixelBuffer & p_buffer)
{
	bool l_bReturn = false;

	if (OpenGl::ReadBuffer( GL_BACK))
	{
		l_bReturn = OpenGl::ReadPixels( 0, 0, m_pGenerator->GetImageSize().x, m_pGenerator->GetImageSize().y, GL_RGBA, GL_UNSIGNED_BYTE, p_buffer.ptr());
	}

	return l_bReturn;
}

void GlFrameBuffer :: Resize( const wxSize & p_size)
{
	m_size = p_size;
	m_bInitialised = false;
}

//*************************************************************************************************

GlFrameVariableBase :: GlFrameVariableBase( GlShaderProgram * p_pProgram, GLuint * p_uiProgram)
	:	m_glIndex( GL_INVALID_INDEX)
	,	m_uiParentProgram( p_uiProgram)
	,	m_bPresentInProgram( true)
	,	m_pProgram( p_pProgram)
	,	m_bChanged( true)
{
}

GlFrameVariableBase :: ~GlFrameVariableBase()
{
}

void GlFrameVariableBase :: GetVariableLocation( const GLchar * p_pVarName)
{
	m_glIndex = OpenGl::GetUniformLocation( * m_uiParentProgram, p_pVarName);
}

//*************************************************************************************************

GLenum GlShaderObject::ShaderTypes[eSHADER_OBJECT_TYPE_COUNT] =
{	GL_VERTEX_SHADER
,	GL_FRAGMENT_SHADER
,	GL_GEOMETRY_SHADER
,	GL_TESS_CONTROL_SHADER
,	GL_TESS_EVALUATION_SHADER
};

GlShaderObject :: GlShaderObject( GlShaderProgram * p_pParent, eSHADER_OBJECT_TYPE p_eType)
	:	m_bCompiled( false)
	,	m_eType( p_eType)
	,	m_bError( false)
	,	m_pParent( p_pParent)
	,	m_shaderObject( GL_INVALID_INDEX)
	,	m_bAttached( false)
{
}

GlShaderObject :: ~GlShaderObject()
{
}

void GlShaderObject :: SetFile( const wxString & p_filename)
{
	m_bError = false;

	m_pathFile.clear();
	m_strSource.clear();

	if ( ! p_filename.empty())
	{
		wxTextFile l_file( p_filename);
		l_file.Open();

		if ( ! l_file.Eof())
		{
			m_pathFile = p_filename;

			for (size_t i = 0 ; i < l_file.GetLineCount() ; i++)
			{
				m_strSource += l_file.GetLine( i);
			}
		}
	}
}

void GlShaderObject :: SetSource( const wxString & p_strSource)
{
	m_bError = false;
	m_strSource = p_strSource;
}

void GlShaderObject :: CreateProgram()
{
	m_shaderObject = OpenGl::CreateShader( ShaderTypes[m_eType]);
}

void GlShaderObject :: DestroyProgram()
{
	Detach();

	if (m_bCompiled && m_shaderObject != GL_INVALID_INDEX)
	{
		OpenGl::DeleteShader( m_shaderObject);
		m_shaderObject = GL_INVALID_INDEX;
		m_bCompiled = false;
	}
}

bool GlShaderObject :: Compile()
{
	bool l_bReturn = false;

	if ( ! m_bError && ! m_strSource.empty())
	{
		m_bCompiled = false;

		GLint l_iCompiled = 0;
		GLint l_iLength = m_strSource.size();
		char * l_pTmp = new char[l_iLength + 1];
		memcpy( l_pTmp, m_strSource.char_str(), l_iLength);
		l_pTmp[l_iLength] = 0;

		OpenGl::ShaderSource( m_shaderObject, 1, const_cast<const char **>( & l_pTmp), & l_iLength);
		OpenGl::CompileShader( m_shaderObject);
		OpenGl::GetShaderiv( m_shaderObject, GL_COMPILE_STATUS, & l_iCompiled);

		delete [] l_pTmp;

		if (l_iCompiled != 0)
		{
			m_bCompiled = true;
		}
		else
		{
			m_bError = true;
		}

		RetrieveCompilerLog( m_compilerLog);

		if (m_bCompiled)
		{
			m_compilerLog.clear();
		}

		l_bReturn = m_bCompiled;
	}

	return l_bReturn;
}

void GlShaderObject :: AttachTo( GlShaderProgram * p_pProgram)
{
	if (m_bCompiled && ! m_bError)
	{
		m_pParent = p_pProgram;
		OpenGl::AttachShader( m_pParent->GetProgramObject(), m_shaderObject);
		m_bAttached = true;
	}
}

void GlShaderObject :: Detach()
{
	if (m_bCompiled && m_pParent != NULL && ! m_bError && m_bAttached)
	{
		try
		{
			OpenGl::DetachShader( m_pParent->GetProgramObject(), m_shaderObject);
			m_bAttached = false;
		}
		catch ( ... )
		{
			wxMessageBox( wxT( "GlShaderObject :: Detach - glDetachShader - Exception"));
		}

		m_pParent = NULL;
		// if you get an error here, you deleted the Program object first and then
		// the ShaderObject! Always delete ShaderObjects last!
	}
}

void GlShaderObject :: RetrieveCompilerLog( wxString & p_strCompilerLog)
{
	int l_iInfologLength = 0;
	int charsWritten  = 0;
	OpenGl::GetShaderiv( m_shaderObject, GL_INFO_LOG_LENGTH, & l_iInfologLength);

	if (l_iInfologLength > 0)
	{
		char * l_szInfoLog = new char[l_iInfologLength];
		OpenGl::GetShaderInfoLog( m_shaderObject, l_iInfologLength, & charsWritten, l_szInfoLog);
		p_strCompilerLog = wxString( l_szInfoLog, wxConvLibc);
		delete [] l_szInfoLog;
	}

	if (p_strCompilerLog.size() > 0)
	{
		p_strCompilerLog = p_strCompilerLog.substr( 0, p_strCompilerLog.size() - 1);
	}
}

//*************************************************************************************************

GlShaderProgram :: GlShaderProgram( const wxArrayString & p_shaderFiles)
	:	m_bLinked		( false)
	,	m_bError		( false)
	,	m_bEnabled		( true)
	,	m_programObject	( 0)
{
	for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT && i < static_cast< int>( p_shaderFiles.size()) ; i++)
	{
		if ( ! p_shaderFiles[i].empty())
		{
			CreateObject( eSHADER_OBJECT_TYPE( i));
			SetProgramFile( p_shaderFiles[i], eSHADER_OBJECT_TYPE( i));
		}
	}
}

GlShaderProgram :: GlShaderProgram()
	:	m_bLinked		( false)
	,	m_bError		( false)
	,	m_bEnabled		( true)
	,	m_programObject	( 0)
{
	for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
	{
		m_pShaders[i] = NULL;
	}
}

GlShaderProgram :: ~GlShaderProgram()
{
	for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
	{
		if (m_pShaders[i])
		{
			delete m_pShaders[i];
			m_pShaders[i] = NULL;
		}
	}

	for (FrameVariablePtrList::iterator it = m_listFrameVariables.begin() ; it != m_listFrameVariables.end() ; ++it)
	{
		delete ( * it);
	}

	m_mapFloatFrameVariables.clear();
	m_mapIntFrameVariables.clear();
	m_listFrameVariables.clear();
}

void GlShaderProgram :: Cleanup()
{
	for (FrameVariablePtrList::iterator it = m_listFrameVariables.begin() ; it != m_listFrameVariables.end() ; ++it)
	{
		delete ( * it);
	}

	m_mapFloatFrameVariables.clear();
	m_mapIntFrameVariables.clear();
	m_listFrameVariables.clear();

	for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
	{
		if (m_pShaders[i] && ! m_pShaders[i]->GetSource().empty())
		{
			m_pShaders[i]->DestroyProgram();
		}
	}

	m_pathFile.clear();

	if (m_programObject != 0)
	{
		OpenGl::DeleteProgram( m_programObject);
	}
}

bool GlShaderProgram :: Initialise()
{
	bool l_bReturn = false;

	if ( ! m_bLinked && ! m_bError)
	{
		Cleanup();

		m_programObject = OpenGl::CreateProgram();

		for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
		{
			if (m_pShaders[i] && ! m_pShaders[i]->GetSource().empty())
			{
				m_pShaders[i]->CreateProgram();
			}
		}

		for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
		{
			if (m_pShaders[i] && ! m_pShaders[i]->GetSource().empty())
			{
				if ( ! m_pShaders[i]->Compile())
				{
					Cleanup();
					m_bError = true;
					return false;
				}
			}
		}

		if ( ! Link())
		{
			Cleanup();
			m_bError = true;
			return false;
		}

		l_bReturn = true;
	}

	return l_bReturn;
}

void GlShaderProgram :: CreateObject( eSHADER_OBJECT_TYPE p_eType)
{
	if ( ! m_pShaders[p_eType])
	{
		m_pShaders[p_eType] = DoCreateObject( p_eType);

		if ( ! m_pathFile.empty())
		{
			m_pShaders[p_eType]->SetFile( m_pathFile);
		}
	}
}

void GlShaderProgram :: DestroyObject( eSHADER_OBJECT_TYPE p_eType)
{
	if (m_pShaders[p_eType])
	{
		m_pShaders[p_eType]->DestroyProgram();
		delete m_pShaders[p_eType];
		m_pShaders[p_eType] = NULL;
	}
}

bool GlShaderProgram :: Link()
{
	bool l_bReturn = false;

	if ( ! m_bError && OpenGl::IsProgram( m_programObject))
	{
		for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
		{
			if (m_pShaders[i] != NULL)
			{
				m_pShaders[i]->AttachTo( this);
			}
		}

		GLint l_iLinked = 0;
		GLint l_iValidated = 0;
		OpenGl::LinkProgram( m_programObject);
		OpenGl::GetProgramiv( m_programObject, GL_LINK_STATUS, & l_iLinked);
		OpenGl::GetProgramiv( m_programObject, GL_VALIDATE_STATUS, & l_iValidated);

		RetrieveLinkerLog( m_linkerLog);

		if (l_iLinked && l_iValidated && m_linkerLog.find( wxT( "ERROR")) == wxString::npos)
		{
			m_bLinked = true;
			m_linkerLog.clear();
		}

		l_bReturn = m_bLinked;
	}

	return l_bReturn;
}

void GlShaderProgram :: RetrieveLinkerLog( wxString & p_strLog)
{
	GLint l_iLength = 0;
	GLsizei l_uiLength = 0;

	if (m_programObject != 0)
	{
		OpenGl::GetProgramiv( m_programObject, GL_INFO_LOG_LENGTH , & l_iLength);

		if (l_iLength > 1)
		{
			char * l_pTmp = new char[l_iLength];
			OpenGl::GetProgramInfoLog( m_programObject, l_iLength, & l_uiLength, l_pTmp);
			p_strLog = wxString( l_pTmp, wxConvLibc);
			delete [] l_pTmp;
		}
	}
}

bool GlShaderProgram :: Activate()
{
	bool l_bReturn = true;

	if (m_programObject != 0 && m_bEnabled && m_bLinked && ! m_bError)
	{
		if (OpenGl::UseProgram( m_programObject))
		{
			l_bReturn = ApplyAllVariables();
		}
	}

	return l_bReturn;
}

void GlShaderProgram :: Deactivate()
{
	if (m_programObject != 0 && m_bEnabled && m_bLinked && ! m_bError)
	{
		OpenGl::UseProgram( 0);
	}
}

void GlShaderProgram :: SetProgramFile( const wxString & p_strFile, eSHADER_OBJECT_TYPE p_eType)
{
	if (m_pShaders[p_eType] != NULL)
	{
		m_pShaders[p_eType]->SetFile( p_strFile);
		ResetToCompile();
	}
}

GlShaderObject * GlShaderProgram :: GetProgram( eSHADER_OBJECT_TYPE p_eType)
{
	return m_pShaders[p_eType];
}

int GlShaderProgram :: GetAttributeLocation( const wxString & p_strName)const
{
	int l_iReturn = GL_INVALID_INDEX;

	if (m_programObject != GL_INVALID_INDEX && OpenGl::IsProgram( m_programObject))
	{
		l_iReturn = OpenGl::GetAttribLocation( m_programObject, p_strName.char_str());
	}

	return l_iReturn;
}

void GlShaderProgram :: SetFile( const wxString & p_path)
{
	m_pathFile = p_path;

	for (int i = eSHADER_OBJECT_TYPE_VERTEX ; i < eSHADER_OBJECT_TYPE_COUNT ; i++)
	{
		if (m_pShaders[i] != NULL)
		{
			m_pShaders[i]->SetFile( p_path);
		}
	}

	ResetToCompile();
}

void GlShaderProgram :: SetProgramText( const wxString & p_strFile, eSHADER_OBJECT_TYPE p_eType)
{
	if (m_pShaders[p_eType] != NULL)
	{
		m_pShaders[p_eType]->SetFile( wxEmptyString);
		m_pShaders[p_eType]->SetSource( p_strFile);
		ResetToCompile();
	}
}

void GlShaderProgram :: ResetToCompile()
{
	m_bLinked = false;
	m_bError = false;
}

bool GlShaderProgram :: ApplyAllVariables()
{
	bool l_bReturn = true;

	for (FrameVariablePtrList::iterator it = m_listFrameVariables.begin() ; it != m_listFrameVariables.end() && l_bReturn ; ++it)
	{
		l_bReturn = ( * it)->Apply();
	}

	return l_bReturn;
}

GlFrameVariable<float> * GlShaderProgram :: CreateFloatFrameVariable( const wxString & p_strName, float p_fValue)
{
	GlFrameVariable<float> * l_pReturn = new GlFrameVariable<float>( this, & m_programObject);
	l_pReturn->SetValue( p_fValue);
	l_pReturn->SetName( p_strName);
	m_listFrameVariables.push_back( l_pReturn);
	m_mapFloatFrameVariables.insert( std::pair<wxString, GlFrameVariable<float> *>( p_strName, l_pReturn));
	return l_pReturn;
}

GlFrameVariable<int> * GlShaderProgram :: CreateIntFrameVariable( const wxString & p_strName, int p_iValue)
{
	GlFrameVariable<int> * l_pReturn = new GlFrameVariable<int>( this, & m_programObject);
	l_pReturn->SetValue( p_iValue);
	l_pReturn->SetName( p_strName);
	m_listFrameVariables.push_back( l_pReturn);
	m_mapIntFrameVariables.insert( std::pair<wxString, GlFrameVariable<int> *>( p_strName, l_pReturn));
	return l_pReturn;
}

void GlShaderProgram :: SetFloatFrameVariableValue( const wxString & p_strName, float p_fValue)
{
	std::map<wxString, GlFrameVariable<float> *>::const_iterator l_it = m_mapFloatFrameVariables.find( p_strName);

	if (l_it != m_mapFloatFrameVariables.end())
	{
		l_it->second->SetValue( p_fValue);
	}
}

void GlShaderProgram :: SetIntFrameVariableValue( const wxString & p_strName, int p_iValue)
{
	std::map<wxString, GlFrameVariable<int> *>::const_iterator l_it = m_mapIntFrameVariables.find( p_strName);

	if (l_it != m_mapIntFrameVariables.end())
	{
		l_it->second->SetValue( p_iValue);
	}
}

GlFrameVariable<int> * GlShaderProgram :: GetIntFrameVariable( const wxString & p_strName)const
{
	GlFrameVariable<int> * l_pReturn = NULL;
	std::map<wxString, GlFrameVariable<int> *>::const_iterator l_it = m_mapIntFrameVariables.find( p_strName);

	if (l_it != m_mapIntFrameVariables.end())
	{
		l_pReturn = l_it->second;
	}

	return l_pReturn;
}

GlFrameVariable<float> * GlShaderProgram :: GetFloatFrameVariable( const wxString & p_strName)const
{
	GlFrameVariable<float> * l_pReturn = NULL;
	std::map<wxString, GlFrameVariable<float> *>::const_iterator l_it = m_mapFloatFrameVariables.find( p_strName);

	if (l_it != m_mapFloatFrameVariables.end())
	{
		l_pReturn = l_it->second;
	}

	return l_pReturn;
}

GlShaderObject * GlShaderProgram :: DoCreateObject( eSHADER_OBJECT_TYPE p_eType)
{
	GlShaderObject * l_pReturn = new GlShaderObject( this, p_eType);
	return l_pReturn;
}

//*************************************************************************************************
