#include "Voronoi2D/GeneratorVoronoi2D.h"
#include <LibGenerator/Random.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;
typedef unsigned char byte;

//*************************************************************************************************
/*
Voronoi2D::Seed :: Seed( int p_iX, int p_iY, UbPixel const & p_pxColour, PixelBuffer * p_pBuffer, SeedPtrMatrix * p_pMatrix)
	:	Point<int, 2>( p_iX, p_iY)
	,	m_pxColour	( p_pxColour)
	,	m_pBuffer	( p_pBuffer)
	,	m_pMatrix	( p_pMatrix)
	,	m_ptMin		( p_iX, p_iY)
	,	m_ptMax		( p_iX, p_iY)
{
}

Voronoi2D::Seed :: Seed( Seed const & p_seed)
	:	Point<int, 2>( p_seed)
	,	m_pxColour	( p_seed.m_pxColour)
	,	m_pBuffer	( p_seed.m_pBuffer)
	,	m_pMatrix	( p_seed.m_pMatrix)
	,	m_ptMin		( p_seed.m_ptMin)
	,	m_ptMax		( p_seed.m_ptMax)
{
}

Voronoi2D::Seed :: ~Seed()
{
}

Voronoi2D::Seed & Voronoi2D::Seed :: operator =( Seed const & p_seed)
{
	Point<int, 2>::operator =( p_seed);
	m_pxColour	= p_seed.m_pxColour;
	m_pBuffer	= p_seed.m_pBuffer;
	m_pMatrix	= p_seed.m_pMatrix;
	m_ptMin		= p_seed.m_ptMin;
	m_ptMax		= p_seed.m_ptMax;
}

void Voronoi2D::Seed :: Grow()
{
	if (m_bCanGrow)
	{
		if (m_ptMin[0] > 0)
		{
			m_ptMin[0]--;
		}
		if (m_ptMin[1] > 0)
		{
			m_ptMin[1]--;
		}
		if (m_ptMax[0] < m_pMatrix->size() - 1)
		{
			m_ptMax[0]++;
		}
		if (m_ptMax[1] < m_pMatrix->at( 0).size() - 1)
		{
			m_ptMax[1]++;
		}
	}
}
*/
//*************************************************************************************************

Voronoi2D::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	ProceduralGenerator::Thread( p_pParent, p_uiIndex, iWidth, iTop, iBottom, iTotalHeight, p_pxColour)
	,	m_pBackBuffer( NULL)
	,	m_pSeeds( NULL)
	,	m_pHeights( NULL)
	,	m_pArrayColours( NULL)
	,	m_pArrayStepColours( NULL)
	,	m_iMaxDepth( 0)
	,	m_iMaxNeighbours( 0)
	,	m_iX( 0)
	,	m_iY( m_iTop)
	,	m_dMkOrder( 4.0)
{
	m_pFunctions[eDISTANCE_FUNC_EUCLIDIAN]			= & Voronoi2D::Thread::EuclidianLength;
	m_pFunctions[eDISTANCE_FUNC_SQUARED_EUCLIDIAN]	= & Voronoi2D::Thread::SquaredEuclidianLength;
	m_pFunctions[eDISTANCE_FUNC_MANHATTAN]			= & Voronoi2D::Thread::ManhattanLength;
	m_pFunctions[eDISTANCE_FUNC_MINKOWSKI]			= & Voronoi2D::Thread::MinkowskiLength;
	m_pFunctions[eDISTANCE_FUNC_CHEBYCHEV]			= & Voronoi2D::Thread::ChebychevLength;
}

Voronoi2D::Thread :: ~Thread()
{
}

void Voronoi2D::Thread :: Step()
{
	if (m_iX < m_iWidth)
	{
		for (int j = m_iTop ; ! IsStopped() && j < m_iBottom ; j++)
		{
			ComputeColour( m_iX, j, ( * m_pBackBuffer)[m_iX][j]);
		}

		m_iX++;
	}
	else
	{
		Stop();
	}
}

void Voronoi2D::Thread :: Initialise( PixelBuffer * p_pBackBuffer, IntSetIntMap * p_pSeeds, IntIntMapIntMap * p_pHeights, ColourArray * p_pArrayColours, ColourArray * p_pArrayStepColours, eDISTANCE_FUNC p_eTypeDistance, int p_iMaxDepth, int p_iMaxNeighbours, double p_dMkOrder)
{
	m_pBackBuffer = p_pBackBuffer;
	m_pSeeds = p_pSeeds;
	m_pHeights = p_pHeights;
	m_pArrayColours = p_pArrayColours;
	m_pArrayStepColours = p_pArrayStepColours;
	m_eTypeDistance = p_eTypeDistance;
	m_iMaxDepth = p_iMaxDepth;
	m_iMaxNeighbours = p_iMaxNeighbours;
	m_dMkOrder = p_dMkOrder;
}

void Voronoi2D::Thread :: ComputeColour( int x, int y, UbPixel & p_px)
{
	Point<int, 2> l_ptCurrent( x, y);
	IntIntMap l_mapHeights;
	IntIntMap::reverse_iterator l_rit;
	int l_iDistance;

	for (IntSetIntMap::const_iterator iti = m_pSeeds->begin() ; ! IsStopped() && iti != m_pSeeds->end() ; ++iti)
	{
		for (IntSet::const_iterator itj = iti->second.begin() ; ! IsStopped() && itj != iti->second.end() ; ++itj)
		{
			l_iDistance = (this->*m_pFunctions[m_eTypeDistance])( l_ptCurrent, Point<int, 2>( iti->first, * itj));
			l_mapHeights.insert( IntIntMap::value_type( l_iDistance, ( * m_pHeights)[iti->first][* itj]));
		}
	}

	while ( ! IsStopped() && l_mapHeights.size() > (size_t)m_iMaxNeighbours)
	{
		l_rit = l_mapHeights.rbegin();
		l_rit++;
		l_mapHeights.erase( l_rit.base());
	}

	int l_iSum = 0;

	for (IntIntMap::iterator l_it = l_mapHeights.begin() ; ! IsStopped() && l_it != l_mapHeights.end() ; ++l_it)
	{
		l_iSum += l_it->second;
	}

	const Colour & l_clr = ( * m_pArrayColours)[l_iSum / m_iMaxNeighbours];
	p_px.Set( l_clr[0], l_clr[1], l_clr[2], byte( 255));
}

int Voronoi2D::Thread :: EuclidianLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB)
{
	return (p_ptB - p_ptA).GetLength();
}

int Voronoi2D::Thread :: SquaredEuclidianLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB)
{
	return (p_ptB - p_ptA).GetSquaredLength();
}

int Voronoi2D::Thread :: ManhattanLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB)
{
	return (p_ptB - p_ptA).GetManhattanLength();
}

int Voronoi2D::Thread :: MinkowskiLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB)
{
	return (p_ptB - p_ptA).GetMinkowskiLength( m_dMkOrder);
}

int Voronoi2D::Thread :: ChebychevLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB)
{
	return (p_ptB - p_ptA).GetChebychevLength();
}

//*************************************************************************************************

Voronoi2D :: Voronoi2D( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_eventType, p_width, p_height, p_iWndId, p_pFrame)
	,	m_iMaxNeighbours( 32)
	,	m_eTypeDistance( eDISTANCE_FUNC_EUCLIDIAN)
	,	m_dMkOrder( 4.0)
{
	m_bNeedsConfigPanel = true;
	Random::srand ( time( NULL));

	m_arrayStepColours.push_back( Voronoi2D::Colour( 000, 000, 000));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 255, 000, 000));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 255, 255, 000));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 000, 255, 000));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 000, 255, 255));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 000, 000, 255));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 255, 000, 255));
	m_arrayStepColours.push_back( Voronoi2D::Colour( 255, 255, 255));
	SetMaxDepth( 5);
}

Voronoi2D :: ~Voronoi2D()
{
	m_initialised = false;

	_cleanup();
}

void Voronoi2D :: SetMaxDepth( int p_iMaxDepth)
{
	Suspend();
	m_iMaxDepth = p_iMaxDepth;

	m_arrayColours.clear();
	int l_iNbColours = m_iMaxDepth * (m_arrayStepColours.size() - 1);
	m_arrayColours.reserve( l_iNbColours);
	int l_iSteps = m_iMaxDepth - 1;
	int l_iCurrent = 0, l_iPrevious;
	Colour l_clrColour;
	Point<double, 3> l_ptStep;
	Point<double, 3> l_ptColour;

	for (int i = 0 ; i < l_iNbColours ;)
	{
		l_iPrevious = l_iCurrent++;
		l_clrColour = m_arrayStepColours[l_iPrevious];
		l_ptColour = m_arrayStepColours[l_iPrevious];
		l_ptStep = (Point<double, 3>( m_arrayStepColours[l_iCurrent]) - l_ptColour) / l_iSteps;

		for (int j = 0 ; j <= l_iSteps ; j++, i++)
		{
			m_arrayColours.push_back( l_ptColour);
			l_ptColour += l_ptStep;
		}
	}

	Initialise();
	Generate();
}

void Voronoi2D :: SetMaxNeighbours( int p_iMaxNeighbours)
{
	Suspend();
	m_iMaxNeighbours = p_iMaxNeighbours;
	Generate();
}

void Voronoi2D :: SetTypeDistance( eDISTANCE_FUNC p_eType)
{
	Suspend();
	m_eTypeDistance = p_eType;
	Generate();
}

void Voronoi2D :: SetMinkowskiOrder( double p_dMkOrder)
{
	Suspend();
	m_dMkOrder = p_dMkOrder;
	Generate();
}

void Voronoi2D :: Initialise()
{
	m_seeds.clear();
	m_heights.clear();

	int l_iNbSeedsX = int( sqrt( float( m_sizeImage.x)));
	int l_iNbSeedsY = int( sqrt( float( m_sizeImage.y)));
	int x, y;
	int l_iIndex;

	IntSetIntMap::iterator l_itSeeds1;
	IntSet::iterator l_itSeeds2;
	IntIntMapIntMap::iterator l_itHeights1;

	for (int i = 0 ; i < l_iNbSeedsX ; i++)
	{
		for (int j = 0 ; j < l_iNbSeedsY ; j++)
		{
			x = Random::rand() % m_sizeImage.x;
			y = Random::rand() % m_sizeImage.y;
			l_itSeeds1 = m_seeds.find( x);
			l_itHeights1 = m_heights.find( x);

			while (l_itSeeds1 != m_seeds.end())
			{
				x = Random::rand() % m_sizeImage.x;
				l_itSeeds1 = m_seeds.find( x);
			}

			m_setX.insert( x);
			m_seeds.insert( IntSetIntMap::value_type( x, IntSet()));
			m_heights.insert( IntIntMapIntMap::value_type( x, IntIntMap()));
			l_itSeeds1 = m_seeds.find( x);
			l_itHeights1 = m_heights.find( x);

			l_itSeeds2 = l_itSeeds1->second.find( y);

			while (l_itSeeds2 != l_itSeeds1->second.end())
			{
				y = Random::rand() % m_sizeImage.y;
				l_itSeeds2 = l_itSeeds1->second.find( y);
			}

			m_setY.insert( y);
			l_itSeeds1->second.insert( y);
			l_iIndex = Random::rand() % m_arrayColours.size();
			l_itHeights1->second.insert( IntIntMap::value_type( y, l_iIndex));
		}
	}
}

void Voronoi2D :: Generate()
{
	int l_iOffset = m_sizeImage.y / m_uiThreadCount;
	Voronoi2D::Thread * l_pThread;

	for (int i = 0 ; i < m_sizeImage.y ; i += l_iOffset)
	{
		l_pThread = CreateThread<Voronoi2D::Thread>( m_sizeImage.x, i, i + l_iOffset, m_sizeImage.y, m_pxColour);
		l_pThread->Initialise( & m_backBuffer, & m_seeds, & m_heights, & m_arrayColours, & m_arrayStepColours, m_eTypeDistance, m_iMaxDepth, m_iMaxNeighbours, m_dMkOrder);
		l_pThread->Run();
	}

	m_initialised = true;
}

//******************************************************************************

IMPLEMENT_PLUGIN( V2dPlugin, V2dPanel, Voronoi2D)

V2dPlugin :: V2dPlugin( wxEventType p_eventType )
	:	Plugin< V2dPanel, Voronoi2D, V2dPlugin >( p_eventType )
{
}

V2dPlugin :: ~V2dPlugin()
{
}

wxString V2dPlugin :: GetName()
{
#if defined( NDEBUG )
	return wxT( "Voronoi 2D");
#else
	return wxT( "Voronoi 2D (Debug)");
#endif
}

//*************************************************************************************************

V2dPanel :: V2dPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
{
}

V2dPanel :: ~V2dPanel()
{
}

void V2dPanel :: Create( Voronoi2D * p_pGenerator, wxWindow * p_pParent, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	Reparent( p_pParent);
	m_pGenerator = p_pGenerator;

	wxString l_strValue;
	new wxStaticText( this, wxID_ANY, wxT( "Max Depth :"), wxPoint( 10, 0), wxSize( p_size.x - 20, 20));
	l_strValue << m_pGenerator->GetMaxDepth();
	m_pStaticMaxDepths = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 90, 0), wxSize( p_size.x - 100, 20));
	m_pSliderMaxDepths = new wxSlider( this, eID_DEPTH, m_pGenerator->GetMaxDepth(), 1, 100, wxPoint( 10, 20), wxSize( p_size.x - 20, 20));

	l_strValue.clear();
	new wxStaticText( this, wxID_ANY, wxT( "Max Neighbours :"), wxPoint( 10, 40), wxSize( p_size.x - 20, 20));
	l_strValue << m_pGenerator->GetMaxNeighbours();
	m_pStaticMaxNeighbours = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 110, 40), wxSize( p_size.x - 110, 20));
	m_pSliderMaxNeighbours = new wxSlider( this, eID_NEIGHBOURS, int( m_pGenerator->GetMaxNeighbours()), 1, 256, wxPoint( 10, 60), wxSize( p_size.x - 20, 20));

	new wxStaticText( this, wxID_ANY, wxT( "Distance Function :"), wxPoint( 10, 90), wxSize( p_size.x - 20, 20));
	wxArrayString l_arrayChoices;
	l_arrayChoices.push_back( wxT( "Euclidian"));
	l_arrayChoices.push_back( wxT( "Squared Euclidian"));
	l_arrayChoices.push_back( wxT( "Manhattan"));
	l_arrayChoices.push_back( wxT( "Minkowski"));
	l_arrayChoices.push_back( wxT( "Chebychev"));
	m_pComboDistanceFunction = new wxComboBox( this, eID_DISTANCE_FUNC, wxT( "Euclidian"), wxPoint( 10, 110), wxSize( p_size.x - 20, wxDEFAULT_HEIGHT), l_arrayChoices, wxCB_READONLY);
	m_pComboDistanceFunction->SetFont( GetFont());

	l_strValue.clear();
	l_strValue << wxT( "Minkowski Order : ") << m_pGenerator->GetMinkowskiOrder();
	m_pStaticMinkowskiOrder = new wxStaticText( this, wxID_ANY, l_strValue, wxPoint( 10, 140), wxSize( p_size.x - 20, 20));
	m_pSliderMinkowskiOrder = new wxSlider( this, eID_MINKOWSKI_ORDER, int( m_pGenerator->GetMinkowskiOrder() * 10), 1, 256, wxPoint( 10, 160), wxSize( p_size.x - 20, 20));
	m_pStaticMinkowskiOrder->Hide();
	m_pSliderMinkowskiOrder->Hide();
}

BEGIN_EVENT_TABLE( V2dPanel, ConfigPanel)
	EVT_COMBOBOX(						eID_DISTANCE_FUNC,		V2dPanel::_onDistanceType)
	EVT_COMMAND_SCROLL_THUMBRELEASE(	eID_NEIGHBOURS,			V2dPanel::_onMaxNeighbours)
	EVT_COMMAND_SCROLL_THUMBRELEASE(	eID_DISTANCE_FUNC,		V2dPanel::_onMaxDepths)
	EVT_COMMAND_SCROLL_THUMBRELEASE(	eID_MINKOWSKI_ORDER,	V2dPanel::_onMkOrder)
	EVT_COMMAND_SCROLL_THUMBTRACK(		eID_NEIGHBOURS,			V2dPanel::_onMaxNeighboursTrack)
	EVT_COMMAND_SCROLL_THUMBTRACK(		eID_DISTANCE_FUNC,		V2dPanel::_onMaxDepthsTrack)
	EVT_COMMAND_SCROLL_THUMBTRACK(		eID_MINKOWSKI_ORDER,	V2dPanel::_onMkOrderTrack)
END_EVENT_TABLE()

void V2dPanel :: _onMaxDepths( wxScrollEvent & p_event)
{
	int l_iValue = p_event.GetInt();

	if (m_pGenerator->IsInitialised())
	{
		m_pGenerator->SetMaxDepth( l_iValue);
		wxString l_strValue;
		l_strValue << l_iValue;
		m_pStaticMaxDepths->SetLabel( l_strValue);
	}
}

void V2dPanel :: _onMaxNeighbours( wxScrollEvent & p_event)
{
	int l_iValue = p_event.GetInt();

	if (m_pGenerator->IsInitialised())
	{
		m_pGenerator->SetMaxNeighbours( l_iValue);
		wxString l_strValue;
		l_strValue << l_iValue;
		m_pStaticMaxNeighbours->SetLabel( l_strValue);
	}
}

void V2dPanel :: _onMkOrder( wxScrollEvent & p_event)
{
	double l_dValue = double( p_event.GetInt()) / 10.0;

	if (m_pGenerator->IsInitialised())
	{
		m_pGenerator->SetMinkowskiOrder( l_dValue);
		wxString l_strValue;
		l_strValue << wxT( "Minkowski Order : ") << l_dValue;
		m_pStaticMinkowskiOrder->SetLabel( l_strValue);
	}
}

void V2dPanel :: _onDistanceType( wxCommandEvent & p_event)
{
	Voronoi2D::eDISTANCE_FUNC l_eValue = Voronoi2D::eDISTANCE_FUNC( p_event.GetInt());

	if (m_pGenerator->IsInitialised())
	{
		m_pGenerator->SetTypeDistance( l_eValue);

		if (l_eValue == Voronoi2D::eDISTANCE_FUNC_MINKOWSKI)
		{
			m_pStaticMinkowskiOrder->Show();
			m_pSliderMinkowskiOrder->Show();
		}
		else
		{
			m_pStaticMinkowskiOrder->Hide();
			m_pSliderMinkowskiOrder->Hide();
		}
	}
}

void V2dPanel :: _onMaxDepthsTrack( wxScrollEvent & p_event)
{
	int l_iValue = p_event.GetInt();
	wxString l_strValue;
	l_strValue << l_iValue;
	m_pStaticMaxDepths->SetLabel( l_strValue);
}

void V2dPanel :: _onMaxNeighboursTrack( wxScrollEvent & p_event)
{
	int l_iValue = p_event.GetInt();
	wxString l_strValue;
	l_strValue << l_iValue;
	m_pStaticMaxNeighbours->SetLabel( l_strValue);
}

void V2dPanel :: _onMkOrderTrack( wxScrollEvent & p_event)
{
	double l_dValue = double( p_event.GetInt()) / 10.0;
	wxString l_strValue;
	l_strValue << wxT( "Minkowski Order : ") << l_dValue;
	m_pStaticMinkowskiOrder->SetLabel( l_strValue);
}

//*************************************************************************************************
